/************************************************************************/
/*                                                                      */
/*   Program Name :  rdf2m5b                                            */
/*                                                                      */
/*     Rdf2m5b is RDF(RadioAstron raw data) to Mark5B file              */
/*     format converter                                                 */
/*                                                                      */
/*                                                                      */
/*            Ver 1.00  2020-09-08 start development based on vdif2m5b  */
/*                                                                      */
/*      Copyright (c) 2020-     T.Kondo All Right Reserved              */
/*                                                                      */
/************************************************************************/
/*


     2020-09-08  -- development started based on vdif2m5b
     2020-09-09  -- Ver 1.0

   how to run
          rdf2m5b rdf_filename [options]
    
     where   rdf_filename -- RDF data file name
              options (any order)

               -fs[ampl] fmhz -- sampling frequency (MHz) 
               -numch  nch    -- # of channels 
               -ad[bit] asbit -- AD resolution 


			   -2bit mode     -- set 2bit AD decode table
			                         1: offset binary   0,1,2,3 (no conversion)
									 2: Mark V          0,2,1,3  (default for Mark5B)
									 3: signed integer  1,0,2,3

               -o m5b_name    -- Mark5B name to be created
                                 (default : see below)
               -d m5b_dir     -- directory for Mark5B file out
                                 (default  : RDF file directory)
               -s[offset] soffset
			         +VE : start offset (sec) (default is 0)
					 -VE : absolute value means skip over bytes 
               -type0   --- naming type 0  rdfname + '.m5b'
               -type1   --- naming type 1  (exchange file extention to 'm5b'

  
  Naming rule for Mark5B file created (RDF.rdf : original VDIF name)
            Type 0:  VDIF.rdf.m5b 
            Type 1:  VDIF.m5b
*/

#define PROGRAM_VERSION "rdf2m5b (Ver 1.00  2020-09-09)"

#include "rdf2m5b.h"

#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory


#define DFLT_DATA_FILE "rdf.dat"
#define BUFMAX 25000   // maximum read buffer size
#define DFLT_CH_TABLE {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16}  // default CH allocation table


int rvsint(int i4)
{
    // reverse integer narabi
    int iw, iwork, ipick, i;
    iw = i4;
    iwork = 0;
    for (i = 0; i < 32; i++) {
        iwork = iwork << 1;
        ipick = iw & 0x1;
        iwork = ipick | (iwork & 0xFFFFFFFE);
        iw = iw >> 1;
    }
    return iwork;
}


int64_t get_file_size(char *fname)
{
    /*  get file size opened */
    int result;
#ifdef WIN32                       //2006-09-20
    struct _stati64 info;    // for VC++    
    result = _stati64(fname, &info);  // for VC++
#else
    struct stat info;
    result = stat(fname, &info);
#endif
    if (result != 0) return (int64_t)(-1);
    else
        return (int64_t)info.st_size;
    // use %lld to print out whole number that has int64_t type
}


int cmnd_get_p(int argc, char *argv[], char *cmnd, char *param)  // 2014-05-22
// do not care about upper case and lower case 

// get argument like -command parameter
// search "-command" in argument precisely (i.e. charcter length must be same), 
// if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];
    char command[80];

    param[0] = '\0';

    // convert to lower case character
    strcpy(command, cmnd);
    conv2small(command);
    //search command
    //printf("%s\n",command); // debug
    n = 2;
    while (n <= argc) {
        strcpy(cbuf, *(argv + n - 1));
        conv2small(cbuf); // convert to small characters
        if (strlen(cbuf) == strlen(command) && strncmp(cbuf, command, strlen(command)) == 0) {
            n++;
            if (n > argc) return 0;
            strcpy(param, *(argv + n - 1));
            return 0;
        }
        n++;
    }
    return -1;
}


void parse_dir_and_name(char *fpath, char *dir, char *fname) //2003-03-23
{
    int len, i, k;
    char cbuf[256];  // 2007-07-17  128->256
    len = strlen(fpath);
    // converting '\' to '/' if so
    for (i = 0; i < len; i++) {
        if (fpath[i] == '\\') {
            cbuf[i] = '/';
        }
        else {
            cbuf[i] = fpath[i];
        }
    }
    cbuf[len] = '\0';
    k = -1;
    for (i = len - 1; i >= 0; i--) {
        if (cbuf[i] == '/') {
            k = i;
            break;
        }
    }
    //printf("debug %d %s\n",k,cbuf);
    if (k < 0) {
        dir[0] = '\0';
        strcpy(fname, cbuf);
    }
    else if (k == len - 1) {
        strcpy(dir, cbuf);
        fname[0] = '\0';
    }
    else {
        strncpy(dir, cbuf, k + 1);
        dir[k + 1] = '\0';
        strcpy(fname, cbuf + k + 1);
    }
    //printf("debug %d %s\n",k,dir);
}


int cmnd_get_num(int argc, char *argv[], char *command, char *param, int *num)
// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];

    param[0]='\0';

    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
			*num=n;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
			*num++;
            return 0;
        }
        n++;
    }
    return -1;
}

int vex_parse(char *cbuf, vparam_t *vparam)  //2003-03-21
{
/*
    Parse VEX parameters by '=',':',';', and ' '.
    '=',':',';' are returned as a parsed item for further analysis
    Total number of items including '=',':',';' is defined in
    type define statesment of vparam as

    typedef struct vparam
    {
        int  n;          //number of param
        char param[60][128];   // 60x128 characters
    } vparam_t;

    In this case 60 is a maximum number of items and
    128 characters are maximum char in an item.

    If runtime error occurrs, please try to increase the number of
    items in the vparam definition.


*/

    // vex_parse =0;  single statement in a cbuf
    //           =1;  more statement left in a cbuf, start point in cbuf 
    //                is updated 
    char abuf[500];
    char *cset=" ";
    char *p=NULL;
    int i,ipos,k,kpos,flag,mflag,len;

    len=strlen(cbuf);
    //printf("len = %d\n",len);

    //remove extra space in front

    if(cbuf[0]==' '){
        kpos=0;
        for(k=0; k<len; k++){
            if(cbuf[k]!=' '){
                kpos=k;
                break;
            }
        }
        i=0;
        for(k=kpos; k<len; k++){
            cbuf[i]=cbuf[k];
            i++;
        }
        cbuf[i]='\0';
    }
    // len again

    len=strlen(cbuf);


    // parse start

    if(cbuf[0] == '*') { // comment
        vparam->n=0; 
        return 0;
    }
    i=0;
    ipos=0;
    flag=0;
    for(k=0; k<len; k++){
        if(cbuf[k] == ' ' || cbuf[k] == '=' || cbuf[k] == ':'){
            if(flag==1) {  // one parameter end
                vparam->param[ipos][i]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            if(cbuf[k] != ' '){
                vparam->param[ipos][i]=cbuf[k];
                vparam->param[ipos][i+1]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            goto nextk;
        }
        if(cbuf[k] == ';'){
            kpos=k;
            goto exithere;
        }
        if(cbuf[k] != ' '){
            flag=1;
            vparam->param[ipos][i]=cbuf[k];
            i++;
        }
nextk:;
    }
    kpos=k;
exithere:
    vparam->param[ipos][i]='\0';
    ipos++;
    vparam->n=ipos;
    i=0;
    mflag=0;
    if(kpos < len-1){
        for(k=kpos+1; k<len; k++) {
            if(cbuf[k] != '\0'){
                abuf[i]=cbuf[k];
                i++;
            }
            if(cbuf[k] == ';') mflag=1;
        }
    }
    if(mflag==1){
        abuf[i]='\0';
        strcpy(cbuf,abuf);
        return 1;
    }
    return 0;
}


void param_get_rdf2m5b(int argc, char *argv[],runp_t *runp)                       
{
    int n,i;
    char cbuf[256];
    int ch_table_dflt[]=DFLT_CH_TABLE;
    vparam_t vp;

	// initialize parameters

    /* common parameters */
    runp->ptype=1;
    runp->fname[0]='\0';
    //runp->k5name[0]='\0';
    runp->m5name[0]='\0';
    runp->m5name_given[0]='\0';
    runp->soffset=0;
    runp->period=0;
    runp->odd=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->unit=0;
    runp->reverse=0;
    //runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    //runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    runp->monit=0;
	runp->flag_compulsory=0;
	runp->skip_bytes=0;
	runp->k5numch=1; // 
    runp->vdifch=1;
    runp->naming_type=0;
    runp->flag_ch_alloc=0;
    runp->pickup_tid=-1;

    /* mode 2 parameters */
    runp->numch=0;
    runp->adbit =0;
    //runp->fmhz  =(double)M5SAMPLE;
    runp->fmhz  =0.0;
	runp->nbitstream=0;
    for(n=0; n<32; n++) runp->ibit[n]=0;
    runp->fmode=3;   // default conversion file mode 1: 1ch  2: 4ch  3: 4x4ch  4: 2x4ch
    //for(n=0; n<4; n++) runp->chtable[n]=0;
    for(n=0; n<16; n++) runp->chtable[n]=ch_table_dflt[n];
    runp->e2bit=2;     // for VDIF -> Mark5B

    //runp->k5nfile=0; //  0 means auto mode  2016-08-08

    /* 1st parameter (vdif file name) get */
    if (argc>=2){
       strcpy(runp->fname,*(argv+1));
	   if (strlen(*(argv+1))==1 && runp->fname[0]=='0'){  // set default name
          strcpy(runp->fname,DFLT_DATA_FILE);
	   }
    }
     /* check VDIFfile name is acompanied by directory or not */
    get_true_path(runp->fname, M5DIR, "./",cbuf);
    strcpy(runp->fname,cbuf);

    /* command parameters get */
    if(cmnd_get_p(argc,argv,"-o",cbuf)==0) strcpy(runp->m5name_given,cbuf);  // Mark5 name
	if(cmnd_get_p(argc,argv,"-d",cbuf)==0){              // out dir
		strcpy(runp->outdir,cbuf);  
        /* env variable is substituted by real one */
        if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
           strcpy(runp->outdir,cbuf);
        }
	}
//    if(cmnd_get(argc,argv,"-c",cbuf)==0) runp->channel = atoi(cbuf);  // CH
//    if(cmnd_get(argc,argv,"-g",cbuf)==0) runp->group   = atoi(cbuf);  // group
    //if(cmnd_get_p(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf);  // start offset
/*
    if(cmnd_get_p(argc,argv,"-p",cbuf)==0) runp->period  = atoi(cbuf);  // period
	if(cmnd_get_p(argc,argv,"-r",cbuf)==0) runp->reverse=1;  // bit reverse mode
	if(cmnd_get_p(argc,argv,"-si",cbuf)==0) strcpy(runp->sid,cbuf);  // station id
	if(cmnd_get_p(argc,argv,"-scan",cbuf)==0) {  // scan
		  runp->scan=atoi(cbuf);
          if(runp->scan < 1) runp->scan=1;
	}
	if(cmnd_get_p(argc,argv,"-monit",cbuf)==0) runp->monit=1;  // monitor mode on
*/
    if(cmnd_get(argc,argv,"-fs",cbuf)==0) runp->fmhz  = atof(cbuf);  // sampling frequency  2018-12-11
    if(cmnd_get_p(argc,argv,"-ad",cbuf)==0) runp->adbit = atoi(cbuf);  // AD bits
    if(cmnd_get_p(argc,argv,"-numch",cbuf)==0) runp->numch = atoi(cbuf);  // # of channels
/*
    if(cmnd_get_p(argc,argv,"-bs",cbuf)==0) runp->nbitstream = atoi(cbuf);  // bit stream

    if(cmnd_get(argc,argv,"-2bit",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2bit encode mode

    if(cmnd_get_p(argc,argv,"-s",cbuf)==0){
		if(atof(cbuf) >= 0.0){
		    runp->soffset = atoi(cbuf);   // second offset
		} else {
			runp->skip_bytes=-atoi(cbuf);  // skip bytes
		}
	}
    if(cmnd_get_p(argc,argv,"-k",cbuf)==0){ // set compulsory mode
		runp->flag_compulsory=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->flag_compulsory=atoi(cbuf);
        }
    }
    if(cmnd_get_p(argc,argv,"-type1",cbuf)==0) runp->naming_type=1;
 */

}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("    rdf2m5b rdf_file [options]\n");
    printf("\n");
    printf("where   rdf_file  -- RDF data file name\n");
    printf("      options (any order)\n");
    printf("        -o m5bname     -- Mark5B file name to be created\n");
    printf("                          (default : see below)\n");
    printf("        -d m5dir      -- directory for Mark5B file out\n");
    printf("                          (default  : same as VDIF file directory)\n");
    //printf("        -bs nbitsream -- set Mark5B # of bit-stream (1,2,4,8,16,32)\n");
    printf("        -fs[ampl] fmhz  -- set sampling frequency (MHz)\n");
    printf("        -ad[bit] adbit  -- AD resolution\n");
    printf("        -numch nch      -- # of channels\n");
    /*
    printf("        -2bit mode    -- set 2bit AD decode table\n");
    printf("                           1: offset binary   0,1,2,3 (no conversion)\n");
    printf("                           2: Mark V standard 0,2,1,3 (default)\n");
    printf("                           3: signed integer  1,0,2,3\n");
    printf("        -s[offset] soffset\n");
    printf("                        +VE : start offset (sec) (default is 0.0)\n");
    */
//    printf("                        -VE : absolute value means skip over bytes \n");
//    printf("        -s strat_sec  -- offset time from data head (sec)\n");
//    printf("                          (default :  0 )\n");
//    printf("        -p period     -- data period to convert (sec)\n");
//    printf("                          (default : all)\n");
//    printf("        -r            -- reverse track order when fanout =2 or more\n");
//    printf("        -v vex_file   -- VEX file name to be read when \"information\n");
//    printf("                              file creation mode\" is set\n");
//    printf("        -sid stat_id  -- station ID when \"information file creation\n");
//    printf("                              mode\" is set\n");
//    printf("        -scan scan#   -- scan# for MODE get when \"information file\n");
//    printf("                              creation mode\" is set (default : 1)\n");
//    printf("        -subp subpass -- set subpass at information file\n");
//    printf("                              creation mode (default is \"A\")\n");
    printf("        -type1        -- outfile naming Type 1 (replace file extention to \"m5b\")\n");

    printf("\n");
    printf("   Naming rule for Mark5B file created (RDF.rdf : original RDF name)\n");
    printf("        Type 0  RDF.rdf.m5b (default)\n");
    printf("        Type 1  RDF.m5b\n");
//    printf("   Enviroment variables\n");
//    printf("        %s    ---   default path to Mark5 data ( %s )\n",
//        M5DIR,getenv(M5DIR));
//    printf("        %s    ---   default path to Mark5 vex file ( %s )\n",
//        M5VEX,getenv(M5VEX));
    printf("\n");
}

void monit_runp(runp_t *runp)
// run parameter monitor
{
       printf(" ---------- RUN CONDITION ---------\n");
       printf("   RDF file name      : %s\n",runp->fname);
       printf("   outdir             : %s\n",runp->outdir);
       //printf("   bitstream : %d\n",runp->nbitstream);
       printf("   sampling freq(MHz) : %f\n",runp->fmhz);
       printf("   AD resolution      : %d\n",runp->adbit);
       printf("   # of channels      : %d\n",runp->numch);
//       printf("   2bit mode : %d\n",runp->e2bit);
//       printf("   infofile  : %s\n",runp->info_file);
//       printf("   channel   : %d\n",runp->channel);
//       printf("   group     : %d\n",runp->group);
//       printf("   soffset   : %d\n",runp->soffset);
//       printf("   period    : %d\n",runp->period);
//       printf("   odd       : %d\n",runp->odd);
//       if(runp->vex_file[0] != '\0')
//          printf("   vex_file  : %s\n",runp->vex_file);
//       if(runp->sid[0] != '\0' && runp->sid[0] !='-')
//          printf("   station_id: %s\n",runp->sid);
       printf(" -------------------------------------\n");
    printf("\n");
}

int main(int argc, char *argv[])
{
       int i;
       double telapsed;           // to measure processing time
       char fname[256];
       char program_version[60];
	   //time_t tstart, tfinish;    
       double celapsed;           // to measure processing time more precisely  2016-08-25
       clock_t cstart, cfinish;   // to measure processing time more precisely  2016-08-25
       runp_t runp;

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       }
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("\n");
          exit(0);
          }
       }

       strcpy(program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_rdf2m5b(argc, argv, &runp); 

       /* out directory create if not set */

       if(runp.outdir[0]=='\0'){
          parse_dir_and_name(runp.fname, runp.outdir, fname); // debug
          //printf("A %s\n",runp.mk5name);
          //printf("A %s\n",runp.outdir);
          //printf("A %s\n",fname);
       }

       //runp.adbit = 1;
       //runp.numch = 4;
       //runp.fmhz = 32.0;

       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    RDF to Mark5B Data Format Converter                   *\n");
       printf("*      %32s  by  T.KONDO       *\n",program_version);
       printf("************************************************************\n");
       printf("\n");

       monit_runp(&runp);

       //tstart= time(NULL);
       cstart = clock();            // 2016-08-25


       i=convRDF2M5B(&runp);

       if(i < 0) goto exitend;

       //printf("  K5 file (%s) has been created.\n\n",runp.k5name);


       //tfinish= time(NULL);
       //telapsed = difftime(tfinish,tstart);
       //printf("\n Time elapsed (sec) for One obs process is %f\n",telapsed);
       printf("\n");
       cfinish = clock();   
       celapsed = (double)(cfinish - cstart) / CLOCKS_PER_SEC; 
       printf("Time elapsed (sec) for conversion is %f\n", celapsed);

exitend: 
       return 0;
}


/* UTC to MJD conversion */
int ut2mjd(double *d, int year, int month, int day,
           int hour, int min, int sec)
/*
  [in]
     year   -- year (yyyy)
     month  -- month (month)
     day    -- day of month (dd)
     hour   -- hour (hh)
     min    -- minute (mm)
     sec    -- second (ss)
  [return]
     ut2mjs -- MJD  (integer day)
     d      -- fractional part of MJD (day)
*/
{
    int  bc, gregory, mnth, y4; 
    //long j;
    int j;

	mnth=month;
	y4=year;
    bc = year <= 0;
    
    /* First day at Gregory's calendar */
    gregory = year > 1582 || (year == 1582 && month > 10) ||
              (year == 1582 && month == 10 && day >= 15);
            
    if (mnth <= 2) {
        y4--;
        mnth += 12;
    }
    
    if (hour < 12) {
        j = 0;
        *d = 0.5;
    } else {
        j = 1;
        *d = -0.5;
    }
    *d += (hour*3600L + min*60 + sec)/86400.0;
    
    j += bc ? (y4-3)/4 : y4/4;
    if (gregory) j+= 2 - y4/100 + y4/400;
    j += 1720994L + y4*365L + (mnth+1)*30 + (mnth+1)*3/5 + day;
    
	/* convert to MJD */
	if (hour < 12) {
		*d -= 0.5;
	} else {
		*d += 0.5;
		j -= 1L;
	}
    j -= 2400000L;
    return j;
}

void ch_realloc(M5B_HEADER_T *m5b, int *chtable)
{
// supposed maximum bitstream 2bit x 16ch
//  Mark5B is 1bit or 2bit AD

    int i,n,m,numch,adbit,bs;
    int pch;
    unsigned int iw, iw32, ow32;
    
    unsigned int imask[4]={0x1, 0x3, 0xF, 0xFF};
    unsigned int umask[4]={0x7FFFFFFF, 0x3FFFFFFF, 0x0FFFFFFF, 0x00FFFFFF};

    adbit=m5b->si.adbit;
    numch=m5b->si.numch;
    bs=adbit*numch;

    // convert each 32-bit word
    for(i=0; i<m5b->ndata; i++){
        iw32=m5b->idata[i];
        ow32=0;
        for(n=0; n< 32/bs; n++){
            for(m=0; m<numch; m++){
                ow32 >>= adbit;
                pch=chtable[m];
                if(pch > numch) pch=m+1;  // doesn't change
                iw= iw32 >> adbit*(pch-1);
                iw &= imask[adbit-1];
                iw <<= 32-adbit;
                ow32= (ow32 & umask[adbit-1]) | iw;
            }
            iw32 >>= bs;
        }
        m5b->idata[i]=ow32;
    }
}



void disp_condition(runp_t *runp, rdf_file_t *rdf)
{
/*
       display RDF and conversion infomation
*/

    int n;
	printf("\n");
	//printf("========================= SUMMARY =============================\n");
	printf("===================== RDF FILE INFO ===========================\n");
	printf(" File Name   : %s\n",runp->fname);
/*
    printf("      Size   : %lld bytes\n",(long long)rdf->f_size); 
	printf(" Start Time  : %04d/%02d/%02d %02d:%02d:%02d\n", 
        vd->vhead.t.year,vd->vhead.t.month,vd->vhead.t.mday,vd->vhead.t.hh,vd->vhead.t.mm,vd->vhead.t.ss);
    printf("               Thread ID : %d      Frame# : %d\n", vd->vhead.thread_id,vd->vhead.frame);
    printf("               Pickup Thread ID : ");
    if(runp->pickup_tid >0){
        printf("%d\n",runp->pickup_tid);
    } else {
        printf("%d\n",vd->vhead.thread_id);
    }
	printf(" Bytes/Frame : %d bytes\n", vd->vhead.frame_len64*8);
	if(vd->vhead.d_type==0){
	    printf(" Data type   : Real Data    ");
    } else {
	    printf(" Data type   : Complex Data ");
    }
    printf("    AD bits : %d", vd->vhead.bpsampl+1);
    printf("    #Channels : %d\n", (int)pow(2.0, vd->vhead.log2_chs));
	printf(" EDV#        : %d\n",vd->vhead.edv);
	if(vd->vhead.edv==1){   // This is NICT's EDV
       printf(" Extended header information (NICT\'s EDV)\n");
	   printf("      Sampling Frequency : %d", vd->vhead.srate);
       if(vd->vhead.uflag==0){
		   printf(" kHz\n");
	   } else {
		   printf(" MHz\n");
	   }
	   printf("      Sync block         : %08X\n", vd->vhead.sync_block);
	   printf("      DAS/Station name   : %s\n", vd->vhead.das);
	}
//	printf(" # of frames/sec     : %d\n",ip->framsec);     
//	printf(" # of total frames   : %d",ip->totalframes);     
//	printf("   (#bad frames : %d)\n",ip->badframes);     
//	printf(" Data period (sec)   : %f\n",ip->dlength);     
//	printf(" Data Rate/thread (Mbps) : %f\n",ip->fbps/1.0e6);     
	printf(" Sampling Frequency Estimated (MHz) : %f\n",vd->fsample/1.0e6); 
    if(runp->flag_ch_alloc==1){
        printf("\n");
        n=0;
        printf(" Channel re-allocation: ");
        while(n < vd->numch){
        //while(n < 16){
            if(n > 0 && (n % 4)==0) printf("\n                        ");            
            printf(" (%2d - %2d)",n+1,runp->chtable[n]);
            n++; 
        }
        printf("\n");
    }
*/

	printf("=================================================================\n");

	printf("\n");

}



int leaf_year_check_t(int year)   // return 0 in case of leaf year
{
    /* leaf year check */
    if (year - (year / 4) * 4 == 0)
    {
        if (year - (year / 100) * 100 == 0)
        {
            if (year - (year / 400) * 400 == 0) { return 0; }
            return -1;
        }
        return 0;
    }
    return -1;
}

void ddd2md_t(int year, int ddd, int* month, int* mday)
{
    int md[12] = { 31,28,31,30,31,30,31,31,30,31,30,31 };
    int i, td, wd;
    /* leaf year check */
    if (leaf_year_check_t(year) == 0)
    {
        md[1] = 29;
    }
    *month = 0;
    *mday = 0;
    td = ddd;
    for (i = 0; i < 12; i++)
    {
        wd = td - md[i];
        if (wd <= 0)
        {
            *mday = td;
            *month = i + 1;
            goto exithere;
        }
        td = wd;
    }
exithere:;
}


void sec2hhmmss_t(data_time_t *t)
{
    t->hh = t->second / 3600;
    t->mm = (t->second - t->hh * 3600) / 60;
    t->ss = t->second % 60;
}

int convRDF2M5B(runp_t *runp)
/*
    VDIF to Mark5B conversion  2016-12-20 
	     (support only 1 ch mode)
*/
{
    int pcount,icount,opos,ipos;
    int i,n,k,m,m4;
    int iwrite;
    int ierr;
	int frame_c;   // total frame numbervcounter 2012-10-17
    int nw;
    unsigned int iw, iw32;
    int64_t totalbyte;
    int64_t tot_readbyte;
    int64_t secbytes;  // # of bytes / sec
    int framesec;  // # of frames /sec
    FILE *fo;
    data_time_t t1st;
    rdf_file_t rdf;
	int *conv2bit;
    int adbit;

	int conv2bit_1[]={0,1,2,3};   // offset binary (no conversion 2 bit)
    int conv2bit_2[]={0,2,1,3};   //  conversion 2 bit (default)
    int conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer
    RDF_HEADER_t *rdfh;

    rdfh = &rdf.rdfh;

    //VDIF_HEADER_T vdif;
    M5B_HEADER_T m5b;
    int seconds_init; // start time in seconds
    int mjd1st;
    double dw,dsec,dmsec10;
    int iread;
    int sec_count;
    int flag_print;



    //printf("  convVDIF2M5B : \n");
    ierr=0;
	frame_c=0; 

    //head=obuf;
    //idata=obuf+4;
    m5b.head=m5b.ibuf;
    m5b.idata=m5b.ibuf+4;
    m5b.nbuf=2504;
    m5b.ndata=2500;

/*
	if(runp->e2bit==1){
	   //ct.conv2bit=conv2bit_1;
	   conv2bit=conv2bit_1;
	} else if(runp->e2bit==3){
	   //ct.conv2bit=conv2bit_3;
	   conv2bit=conv2bit_3;
	} else {
	   //ct.conv2bit=conv2bit_2;
	   conv2bit=conv2bit_2;
	}
*/

	strcpy(rdf.fname,runp->fname);
//	vd.skip_bytes=runp->skip_bytes;
//	vd.flag_compulsory=runp->flag_compulsory;

    /* get RDF file info by opening and get reading file (then close) */
    if(rdf_t_ini(&rdf)<0) exit(0);   // RDF data file header read

    monit_rdf_header(&rdf);

    if(runp->fmhz==0.0){
       printf("Warning!  sampling frequency is not included in VDIF headder!!\n");
       printf("Please set sampling frequency using \"-fsampl fMHz\" option!!\n");
       exit(0);
    }
    else {
        m5b.fsample = 1.0e6*runp->fmhz;
    }


//    disp_condition(runp, &vd);

    
    t1st.year = rdfh->itime[0];
    t1st.ddd = rdfh->itime[1];
    t1st.hh = rdfh->itime[2];
    t1st.mm = rdfh->itime[3];
    t1st.ss = rdfh->itime[4];
    ddd2md_t(t1st.year, t1st.ddd, &t1st.month, &t1st.mday);
    t1st.second = t1st.hh * 3600 + t1st.mm * 60 + t1st.ss;

    printf("  RDF file start time :  %4d/%02d/%02d (%03d) %02d:%02d:%02d\n",
        t1st.year, t1st.month, t1st.mday, t1st.ddd, t1st.hh, t1st.mm, t1st.ss);


    if(rdf.flag_open==0){ 
	    // RDF file open for conversion 
	    if( (rdf.fi  = fopen( rdf.fname, "rb" )) == NULL ){
             printf("File (%s) Open Error! \n",rdf.fname);
             goto exit2return;
        }
	    rdf.flag_open=1;
        // pointer move to the beginning of data area
        rdf.skip_bytes = 256;
        fseek(rdf.fi, rdf.skip_bytes, SEEK_SET);

    }
    // foward startoffset seconds 
/*
	if(runp->soffset>0){
		printf("  %d second forwading.....",runp->soffset);
	    if(fwdNsecVDIF(runp->soffset, &vd) <0){
           printf("Foward error %d\n",runp->soffset);
           i=fclose( vd.fin ) ;
           goto exit2return;
        }
	    printf("\n");
	    // get headder info again 
        if(vdif_header_read_mode1(&vd)<0) goto exit2return;
	    time1st=vd.vhead.t;
	}
*/

    // output file create 
    if(runp->m5name[0]=='\0'){
       m5bname_make(runp);
    }
    printf("\n");
    printf(" Mark5B Data File will be created : %s\n",runp->m5name);
    if( (fo  = fopen( runp->m5name, "wb" )) == NULL ){
           printf("      File  %s  Open Error! \n",runp->m5name);
           ierr= -2;
           goto exit2return;
    }

    pcount=0; // converted data period counter (in sec)

    // header monitor 

    printf("  ****************************************************************\n");
    printf("               Mark5B Header Data\n");
    printf("                                TIME CODE  CRC-  \n");
    printf("   HH:MM:SS   SYNC   UUUUTFFF MJDsssss.ssss  16  DATA#1   DATA#2 \n");
    printf("  ----------------------------------------------------------------\n");

    // Mark5 header common info setup 

    m5b.t.mjd=ut2mjd(&dw,t1st.year,t1st.month,t1st.mday,0,0,0);
    m5b.t.mjd=m5b.t.mjd % 1000;
    mjd1st=m5b.t.mjd;
    m5b.t.msec10=0;
    seconds_init=t1st.second;
    m5b.t.second=seconds_init;
    m5b.tvg=0;
    m5b.uid=0xF00F;
    dsec=0.0;
    dmsec10=0.0;
    sec_count=0;
    m5b.frame=0;  // frame # (in a second) counter reset
    sec2hhmmss_t(&m5b.t);


    //m5b.fsample=vd.fsample;
    m5b.p_frame=(double)((m5b.ndata*32)/(runp->adbit*runp->numch))/m5b.fsample;  // period/frame
    if(((m5b.ndata*32) % (runp->adbit*runp->numch)) !=0 ){
        printf("conRDF2M5B:  Error!!  (cannot convert to Mark5B format)\n");
        exit(0);
    }
    framesec = (int)(1.0 / m5b.p_frame + 0.5);
    adbit=runp->adbit;

    m5b.si.adbit=runp->adbit;
    m5b.si.numch=runp->numch;
    m5b.si.sfkHz=(int)(m5b.fsample/1.0e3 + 0.5);

    secbytes = m5b.fsample*runp->adbit*runp->numch / 8;  // bytes/sec

    // Mark5B initial header create 
       // Mark5B header generation 
    make_m5b_header(&m5b);


    totalbyte=0;
    flag_print = 1;
    m5b.frame = 0;

    while (1) {

        // frame loop
        iread = fread(m5b.idata, sizeof(int), m5b.ndata, rdf.fi);
        if (iread == 0) { // EOF
            fclose(rdf.fi);
            i = fclose(rdf.fi);
            goto exit2return;
        }

        if (flag_print == 1) {
            //printf("   %02d:%02d:%02d %6d", m5b.t.hh, m5b.t.mm, m5b.t.ss, m5b.frame);
            printf("   %02d:%02d:%02d ", m5b.t.hh, m5b.t.mm, m5b.t.ss);
            for (i = 0; i < 4; i++) {
                //printf("%08X ",rvsint(head[i]));
                printf("%08X ", m5b.head[i]);
            }
            for (i = 0; i < 2; i++) {
                printf("%08X ", m5b.idata[i]);
            }
            printf("\n");
            flag_print = 0;
        }

        iwrite = fwrite(m5b.ibuf, sizeof(int), m5b.nbuf, fo);  // write one frame data including header
        totalbyte += m5b.nbuf * sizeof(int);
        m5b.frame++;
        if (m5b.frame >= framesec) {
            flag_print = 1;
            m5b.frame = 0;
            sec_count++;
        }

        // time update for Mark5B header
        m5b_time_update(&m5b, mjd1st, seconds_init, sec_count);
        make_m5b_header(&m5b);    // header update
        sec2hhmmss_t(&m5b.t);

    }


exit2return:
    if(rdf.flag_open==1) i=fclose(rdf.fi);
    printf("  ****************************************************************\n");
    printf("\n");
    printf(" Mark5B Data File created : %s\n",runp->m5name);
    if(totalbyte ==0){
        printf("                             but contents are empty!!\n");
    } else {
        printf("                            (%lld bytes)\n",(long long)totalbyte);
    }

    return 0;
}


void m5bname_make(runp_t *rp)
{
/*
  Create M5b file name  
         2016-12-20
*/
    int i,len;
    //int flag_remove=0;
    int flag_remove;
    char m5dir[256],dir[256], name[256],fname[256]; 
    char dir2[256],name2[256];

    flag_remove=0;
    if(rp->naming_type==1) flag_remove=1;

    //parse_dir_and_name(adsname, dir, name);
    parse_dir_and_name(rp->fname, dir, name);

    /* remove file extension if set so */

    if(flag_remove==1){
        len=strlen(name);

         // find last "." in the name to remove extension
        for(i=len-1; i>=0; i--){
           if(name[i]=='.'){
               name[i]='\0';
               break;
           }
        }
    }


    // out directory prepare
    if(rp->outdir[0] == '\0'){
        m5dir[0]='\0';
    } else {
        // converting '\' to '/' if so
        len=strlen(rp->outdir);
        //printf("C %d %s\n",len,outdir);
        for(i=0; i<len; i++){
            if(rp->outdir[i]=='\\'){
                m5dir[i]='/';
            } else {
                m5dir[i]=rp->outdir[i];
            }
        }
        m5dir[len]='\0';
        // checkdir name accompanied by "/" or not 
        if(m5dir[len-1] != '/') {
           m5dir[len]='/';
           m5dir[len+1]='\0';
        }
    }

    // check given m5name is accompanied by directory or not
    dir2[0]='\0'; name2[0]='\0';
    if(rp->m5name_given[0] != '\0'){
        parse_dir_and_name(rp->m5name_given, dir2, name2);
    }

    // create m5bname

    if(rp->m5name_given[0] == '\0'){  // in case ofnot given from an operator
       strcpy(fname,name);
       strcat(fname,".m5b");
       strcpy(rp->m5name,m5dir);
       strcat(rp->m5name,fname);
    } else {   // in case of given from an operator
        if(dir2[0]=='\0'){
            if(rp->outdir[0]=='\0'){ 
                strcpy(rp->m5name,rp->m5name_given);
            } else {
                strcpy(rp->m5name,m5dir);
                strcat(rp->m5name,name2);
            }
        } else {
            strcpy(rp->m5name,rp->m5name_given);
        }
    }
}


//void m5b_hd_make(int *head, int mjd, int seconds, int msec10,
//				 int uid, int tvgflag, int nframe)
void make_m5b_header(M5B_HEADER_T *m5b)
/*
     Mark5B header block creation
     2009-06-07 

  [in]
      mjd      --  MJD (3digits)
	  seconds  --- seconds in day
	  msec10   --- msec*10
	  uid      --   User-specified field  (lower 16-bits are valid)
	  tvgflag  --   Tvg data flag   1: tvg data  0: not tvg data
	  nframe   --   Frame# within second (starting at 0)]
  [out]
      head  ---- pointer to header array (4 X 32 bits) buffer 
*/
{
    /* Mark 5B header block creation */
    int n,a;
    int crc16;
    //int64_t lw1;
    int idat[12];
    unsigned char bdat[6];
    int sync_pat=0xABADDEED;   // Mark5B sync pattern


    /* Sync word   */
	m5b->head[0]=sync_pat;

	/* User-specfied field (16 bits) */

	m5b->head[1]=(m5b->uid & 0xFFFF) << 16;

	/* TVG flag */
	m5b->head[1]=((m5b->head[1] & 0xFFFF0000) | ((m5b->tvg & 0x1) << 15));

	/* Frame # field */
	//head[1]=((head[1] & 0xFFFF8000) | (nframe & 0x7FFF));
	m5b->head[1]=((m5b->head[1] & 0xFFFF8000) | (m5b->frame & 0x7FFF));

    /* Time field */

    //  JJJSSSSSssss

    a=100;
    for(n=0; n<3; n++){
        idat[n]=(m5b->t.mjd/a) % 10;
        a/=10;
    }
    a=10000;
    for(n=3; n<8; n++){
        idat[n]=(m5b->t.second/a) % 10;
        a/=10;
    }
    a=1000;
    for(n=8; n<12; n++){
        idat[n]=(m5b->t.msec10/a) % 10;
        a/=10;
    }

	m5b->head[2]=idat[0] & 0xF;
    for(n=1; n<8; n++){  // 4 bit step
		m5b->head[2]=((m5b->head[2] << 4) | (idat[n] & 0xF));
	}

	m5b->head[3]=idat[8] & 0xF;
    for(n=9; n<12; n++){  // 4 bit step
		m5b->head[3]=((m5b->head[3] << 4) | (idat[n] & 0xF));
	}
	m5b->head[3] <<= 16;

    /* CRC16 */   
    for(n=0; n<6; n++){
        bdat[n]= (idat[2*n] << 4) + idat[2*n+1];
    }
	crc16 = crcc(bdat, 48, 040003, 16);
    //printf("Debug CRC16 %08X\n",rvsint(crc16));
    //printf("Debug CRC16 %08X\n",crc16);

	// Mark5B bit order is reversal 
	crc16=rvsint(crc16) >> 16;


	m5b->head[3] |= (crc16 & 0xFFFF);
    
}



/* crc subroutine ripoff  by David*/
int crcc(unsigned char* idata, int len, int mask, int cycl) {
	unsigned int istate = 0;
	int idbit;
	int q, ich, icb;

	for (idbit = 1; idbit <= len; idbit++) {
		q = istate & 1;
		ich = (idbit - 1) / 8;
		icb = 7 - (idbit - 1) % 8;
		if ((((idata[ich] >> icb) & 1) ^ q) == 0)
			istate &= -2;
		else {
			istate ^= mask;
			istate |= 1;
		}
		istate = (istate >> 1) | ((istate & 1) << (cycl -1));  
	}

	return ((int) istate);
}


void m5b_time_update(M5B_HEADER_T *m5b, int mjd1st, int seconds_init, int sec_count)
{
   // time update for Mark5B 
    double dsec,dmsec10;
    dsec=(double)m5b->frame*m5b->p_frame;   // fractional second
    dmsec10=dsec*10000.0;
    m5b->t.msec10=(int)dmsec10;
    m5b->t.second=seconds_init+sec_count;
    m5b->t.mjd=mjd1st+(int)((double)m5b->t.second/86400.0);
    if(m5b->t.second >= 86400) m5b->t.second-=86400;
}



int rdf_t_ini(rdf_file_t *rdf)
{
    /*
         read RDF file header
         2020-09-09

         return  0: OK
                -1: error occurred
    */

    int n, nofst;
    int iread;
    int headlen = 256;
    RDF_HEADER_t *rdfh;
    char sbuf[10];

    rdfh = &rdf->rdfh;
    rdf->flag_open = 0;
    // file open
    if ((rdf->fi = fopen(rdf->fname, "rb")) == NULL) {
        printf("rdf_t_ini: File (%s) Open Error! \n", rdf->fname);
        return -99;
    }
    rdf->flag_open = 1;
    rdf->pos = 0;
    // file size get
    rdf->f_size = get_file_size(rdf->fname);

//    if (rdf->skip_bytes > 0) {  // skip 1st skip_byte
//        printf(" get_real_vdif_info: Skipping 1st %lld bytes .....", (long long)rdf->skip_bytes);
//        fseek(rdf->fi, rdf->skip_bytes, SEEK_SET);
//        printf("\n");
//    }

    // read header
    iread = fread(rdf->head, sizeof(unsigned char), headlen, rdf->fi);
    if (iread == 0) { // EOF
        n=fclose(rdf->fi);
        return -1;
    }
    //rdf->pos = ftell(rdf->fi);  //  current pointer position

    // parse header
    nofst = 0;
    memcpy(rdfh->async,   rdf->head + nofst,  4); nofst +=  4; rdfh->async[4] = '\0';
    memcpy(&rdfh->hsize,  rdf->head + nofst,  2); nofst +=  2; 
    memcpy(rdfh->autc,    rdf->head + nofst, 18); nofst += 18; rdfh->autc[18] = '\0';
    memcpy(rdfh->station, rdf->head + nofst, 11); nofst += 11; rdfh->station[11] = '\0';
    memcpy(rdfh->source,  rdf->head + nofst, 11); nofst += 11; rdfh->source[11] = '\0';
    memcpy(rdfh->exp,     rdf->head + nofst, 11); nofst += 11; rdfh->exp[11] = '\0';
    memcpy(rdfh->adatrate,rdf->head + nofst,  2); nofst +=  2; rdfh->adatrate[2] = '\0';
    memcpy(&rdfh->armode, rdf->head + nofst,  1); nofst +=  1;
    memcpy(rdfh->dmode,   rdf->head + nofst,  3); nofst +=  3; rdfh->dmode[3] = '\0';
    memcpy(&rdfh->x,      rdf->head + nofst,  8); nofst +=  8;
    memcpy(&rdfh->y,      rdf->head + nofst,  8); nofst +=  8;
    memcpy(&rdfh->z,      rdf->head + nofst,  8); nofst +=  8;
    memcpy(&rdfh->s_hr,   rdf->head + nofst,  8); nofst +=  8;
    memcpy(&rdfh->s_deg,  rdf->head + nofst,  8); nofst +=  8;
    memcpy(&rdfh->delay,  rdf->head + nofst,  4); nofst +=  4;
    memcpy(&rdfh->qt,     rdf->head + nofst,  1); nofst +=  1;
    memcpy(&rdfh->nch,    rdf->head + nofst,  1); nofst +=  1;
    memcpy(rdfh->reserve, rdf->head + nofst, 51); nofst += 51;
    for (n = 0; n < 16; n++) {
        memcpy(&rdfh->chdat[n].bpol,   rdf->head + nofst, 1); nofst += 1;
        memcpy(&rdfh->chdat[n].bstrip, rdf->head + nofst, 1); nofst += 1;
        memcpy(&rdfh->chdat[n].freq,   rdf->head + nofst, 4); nofst += 4;
    }
    fclose(rdf->fi);
    rdf->flag_open = 0;
    rdf->pos = 0;

    rdfh->d_Brate = atof(rdfh->adatrate);
    strncpy(sbuf, rdfh->autc,    4); sbuf[4] = '\0';  rdfh->itime[0] = atoi(sbuf);   //  YYYY
    strncpy(sbuf, rdfh->autc+5,  3); sbuf[3] = '\0';  rdfh->itime[1] = atoi(sbuf);   //  DDD
    strncpy(sbuf, rdfh->autc+9,  2); sbuf[2] = '\0';  rdfh->itime[2] = atoi(sbuf);   //  HH
    strncpy(sbuf, rdfh->autc+12, 2); sbuf[2] = '\0';  rdfh->itime[3] = atoi(sbuf);   //  MM
    strncpy(sbuf, rdfh->autc+15, 2); sbuf[2] = '\0';  rdfh->itime[4] = atoi(sbuf);   //  SS

    return 0;
}


void monit_rdf_header(rdf_file_t *rdf)
{

    int n;
    RDF_HEADER_t *rdfh;

    rdfh = &rdf->rdfh;

    printf("===================== RDF FILE INFO ===========================\n");
    printf("  Sync word            ;  %4s\n", rdfh->async);
    printf("  Header size          ;  %d\n", rdfh->hsize);
    printf("  UTC                  ;  %s  (%4d/%03d %02d:%02d:%02d)\n", rdfh->autc,
        rdfh->itime[0],rdfh->itime[1],rdfh->itime[2],rdfh->itime[3],rdfh->itime[4]);
    printf("  Telescope name       ;  %s\n", rdfh->station);
    printf("  Source name          ;  %s\n", rdfh->source);
    printf("  Experiment           ;  %s\n", rdfh->exp);
    printf("  Total Stream in MB/s ;  %s  (%f)\n", rdfh->adatrate, rdfh->d_Brate);
    printf("  Reception mode       ;  %c\n", rdfh->armode);
    
    printf("  DDR mode             ;  %s\n", rdfh->dmode);
    printf("===============================================================\n");

    /*
    for (n = 0; n < 16; n++) {
        printf(" CH# %2d\n", n + 1);
        printf("     Polarization --  %d\n", (int)rdfh->chdat[n].bpol);
        printf("     Strip        --  %d\n", (int)rdfh->chdat[n].bstrip);
        printf("     Frequency    --  %g\n", rdfh->chdat[n].freq);
    }
    */
}