/*

   FFT package  Ver.2017-02-16


   FFT package for limited data length (power of 2)
   and non-limited data lenght(any number)


   void cdft(int n, int isgn, double *a)   for data length is power of 2

   void fft1d_mr(double *cr, double *ci, int nfft, int isgnn)  for any data length


   in case of forward FFT (time->frequency),  normalization by NFFT is already 
   carried out inside these functions


References

   "cdft"


   "fft1d_mr"  Mixed-Radix method

   Singleton, R. C., An algorithm for computing the Mixed Radix Fast Fourier Transform,
      IEEE Trans. Audio and Electroacoustic, Vol. AU-17, No.2, pp.93-103, 1969.


*/


#include "libfft.h"

#include <math.h>

#ifndef PIRA
#define PIRA  1.745329252E-2    /* PI/180  degree to radian */
#endif

#ifndef M_PI_2
#define M_PI_2      1.570796326794896619231321691639751442098584699687
#endif
#ifndef WR5000  /* cos(M_PI_2*0.5000) */
#define WR5000      0.707106781186547524400844362104849039284835937688
#endif
#ifndef WR2500  /* cos(M_PI_2*0.2500) */
#define WR2500      0.923879532511286756128183189396788286822416625863
#endif
#ifndef WI2500  /* sin(M_PI_2*0.2500) */
#define WI2500      0.382683432365089771728459984030398866761344562485
#endif


#ifndef RDFT_LOOP_DIV  /* control of the RDFT's speed & tolerance */
#define RDFT_LOOP_DIV 64
#endif

#ifndef DCST_LOOP_DIV  /* control of the DCT,DST's speed & tolerance */
#define DCST_LOOP_DIV 64
#endif


/*
Fast Fourier/Cosine/Sine Transform
    dimension   :one
    data length :power of 2
    decimation  :frequency
    radix       :4, 2
    data        :inplace
    table       :not use
functions
    cdft: Complex Discrete Fourier Transform
    rdft: Real Discrete Fourier Transform
    ddct: Discrete Cosine Transform
    ddst: Discrete Sine Transform
    dfct: Cosine Transform of RDFT (Real Symmetric DFT)
    dfst: Sine Transform of RDFT (Real Anti-symmetric DFT)
function prototypes
    void cdft(int, int, double *);
    void rdft(int, int, double *);
    void ddct(int, int, double *);
    void ddst(int, int, double *);
    void dfct(int, double *);
    void dfst(int, double *);


-------- Complex DFT (Discrete Fourier Transform) --------
    [definition]
        <case1>
            X[k] = sum_j=0^n-1 x[j]*exp(2*pi*i*j*k/n), 0<=k<n
        <case2>
            X[k] = sum_j=0^n-1 x[j]*exp(-2*pi*i*j*k/n), 0<=k<n
        (notes: sum_j=0^n-1 is a summation from j=0 to n-1)
    [usage]
        <case1>
            cdft(2*n, 1, a);
        <case2>
            cdft(2*n, -1, a);
    [parameters]
        2*n            :data length (int)
                        n >= 1, n = power of 2
        a[0...2*n-1]   :input/output data (double *)
                        input data
                            a[2*j] = Re(x[j]), 
                            a[2*j+1] = Im(x[j]), 0<=j<n
                        output data
                            a[2*k] = Re(X[k]), 
                            a[2*k+1] = Im(X[k]), 0<=k<n
    [remark]
        Inverse of 
            cdft(2*n, -1, a);
        is 
            cdft(2*n, 1, a);
            for (j = 0; j <= 2 * n - 1; j++) {
                a[j] *= 1.0 / n;
            }
        .


-------- Real DFT / Inverse of Real DFT --------
    [definition]
        <case1> RDFT
            R[k] = sum_j=0^n-1 a[j]*cos(2*pi*j*k/n), 0<=k<=n/2
            I[k] = sum_j=0^n-1 a[j]*sin(2*pi*j*k/n), 0<k<n/2
        <case2> IRDFT (excluding scale)
            a[k] = (R[0] + R[n/2]*cos(pi*k))/2 + 
                   sum_j=1^n/2-1 R[j]*cos(2*pi*j*k/n) + 
                   sum_j=1^n/2-1 I[j]*sin(2*pi*j*k/n), 0<=k<n
    [usage]
        <case1>
            rdft(n, 1, a);
        <case2>
            rdft(n, -1, a);
    [parameters]
        n              :data length (int)
                        n >= 2, n = power of 2
        a[0...n-1]     :input/output data (double *)
                        <case1>
                            output data
                                a[2*k] = R[k], 0<=k<n/2
                                a[2*k+1] = I[k], 0<k<n/2
                                a[1] = R[n/2]
                        <case2>
                            input data
                                a[2*j] = R[j], 0<=j<n/2
                                a[2*j+1] = I[j], 0<j<n/2
                                a[1] = R[n/2]
    [remark]
        Inverse of 
            rdft(n, 1, a);
        is 
            rdft(n, -1, a);
            for (j = 0; j <= n - 1; j++) {
                a[j] *= 2.0 / n;
            }
        .


-------- DCT (Discrete Cosine Transform) / Inverse of DCT --------
    [definition]
        <case1> IDCT (excluding scale)
            C[k] = sum_j=0^n-1 a[j]*cos(pi*j*(k+1/2)/n), 0<=k<n
        <case2> DCT
            C[k] = sum_j=0^n-1 a[j]*cos(pi*(j+1/2)*k/n), 0<=k<n
    [usage]
        <case1>
            ddct(n, 1, a);
        <case2>
            ddct(n, -1, a);
    [parameters]
        n              :data length (int)
                        n >= 2, n = power of 2
        a[0...n-1]     :input/output data (double *)
                        output data
                            a[k] = C[k], 0<=k<n
    [remark]
        Inverse of 
            ddct(n, -1, a);
        is 
            a[0] *= 0.5;
            ddct(n, 1, a);
            for (j = 0; j <= n - 1; j++) {
                a[j] *= 2.0 / n;
            }
        .


-------- DST (Discrete Sine Transform) / Inverse of DST --------
    [definition]
        <case1> IDST (excluding scale)
            S[k] = sum_j=1^n A[j]*sin(pi*j*(k+1/2)/n), 0<=k<n
        <case2> DST
            S[k] = sum_j=0^n-1 a[j]*sin(pi*(j+1/2)*k/n), 0<k<=n
    [usage]
        <case1>
            ddst(n, 1, a);
        <case2>
            ddst(n, -1, a);
    [parameters]
        n              :data length (int)
                        n >= 2, n = power of 2
        a[0...n-1]     :input/output data (double *)
                        <case1>
                            input data
                                a[j] = A[j], 0<j<n
                                a[0] = A[n]
                            output data
                                a[k] = S[k], 0<=k<n
                        <case2>
                            output data
                                a[k] = S[k], 0<k<n
                                a[0] = S[n]
    [remark]
        Inverse of 
            ddst(n, -1, a);
        is 
            a[0] *= 0.5;
            ddst(n, 1, a);
            for (j = 0; j <= n - 1; j++) {
                a[j] *= 2.0 / n;
            }
        .


-------- Cosine Transform of RDFT (Real Symmetric DFT) --------
    [definition]
        C[k] = sum_j=0^n a[j]*cos(pi*j*k/n), 0<=k<=n
    [usage]
        dfct(n, a);
    [parameters]
        n              :data length - 1 (int)
                        n >= 2, n = power of 2
        a[0...n]       :input/output data (double *)
                        output data
                            a[k] = C[k], 0<=k<=n
    [remark]
        Inverse of 
            a[0] *= 0.5;
            a[n] *= 0.5;
            dfct(n, a);
        is 
            a[0] *= 0.5;
            a[n] *= 0.5;
            dfct(n, a);
            for (j = 0; j <= n; j++) {
                a[j] *= 2.0 / n;
            }
        .


-------- Sine Transform of RDFT (Real Anti-symmetric DFT) --------
    [definition]
        S[k] = sum_j=1^n-1 a[j]*sin(pi*j*k/n), 0<k<n
    [usage]
        dfst(n, a);
    [parameters]
        n              :data length + 1 (int)
                        n >= 2, n = power of 2
        a[0...n-1]     :input/output data (double *)
                        output data
                            a[k] = S[k], 0<k<n
                        (a[0] is used for work area)
    [remark]
        Inverse of 
            dfst(n, a);
        is 
            dfst(n, a);
            for (j = 1; j <= n - 1; j++) {
                a[j] *= 2.0 / n;
            }
        .
*/


void cdft(int n, int isgn, double *a)
{
    void bitrv2(int n, double *a);
    void bitrv2conj(int n, double *a);
    void cftfsub(int n, double *a);
    void cftbsub(int n, double *a);
    int j;
    if (n > 4) {
        if (isgn >= 0) {
            bitrv2(n, a);
            cftfsub(n, a);
        } else {
            bitrv2conj(n, a);
            cftbsub(n, a);
        }
    } else if (n == 4) {
        cftfsub(n, a);
    }
	/* following is to keep compatibility with PV-WAVE at forward (time->frequency) FFT */
	if (isgn < 0){
		for (j = 0; j <= 2 * (n/2) - 1; j++) {
           a[j] *= 1.0 / (n/2);
		}	
	}
}


void rdft(int n, int isgn, double *a)
{
    void bitrv2(int n, double *a);
    void cftfsub(int n, double *a);
    void cftbsub(int n, double *a);
    void rftfsub(int n, double *a);
    void rftbsub(int n, double *a);
    double xi;
    
    if (isgn >= 0) {
        if (n > 4) {
            bitrv2(n, a);
            cftfsub(n, a);
            rftfsub(n, a);
        } else if (n == 4) {
            cftfsub(n, a);
        }
        xi = a[0] - a[1];
        a[0] += a[1];
        a[1] = xi;
    } else {
        a[1] = 0.5 * (a[0] - a[1]);
        a[0] -= a[1];
        if (n > 4) {
            rftbsub(n, a);
            bitrv2(n, a);
            cftbsub(n, a);
        } else if (n == 4) {
            cftfsub(n, a);
        }
    }
}


void ddct(int n, int isgn, double *a)
{
    void bitrv2(int n, double *a);
    void cftfsub(int n, double *a);
    void cftbsub(int n, double *a);
    void rftfsub(int n, double *a);
    void rftbsub(int n, double *a);
    void dctsub(int n, double *a);
    void dctsub4(int n, double *a);
    int j;
    double xr;
    
    if (isgn < 0) {
        xr = a[n - 1];
        for (j = n - 2; j >= 2; j -= 2) {
            a[j + 1] = a[j] - a[j - 1];
            a[j] += a[j - 1];
        }
        a[1] = a[0] - xr;
        a[0] += xr;
        if (n > 4) {
            rftbsub(n, a);
            bitrv2(n, a);
            cftbsub(n, a);
        } else if (n == 4) {
            cftfsub(n, a);
        }
    }
    if (n > 4) {
        dctsub(n, a);
    } else {
        dctsub4(n, a);
    }
    if (isgn >= 0) {
        if (n > 4) {
            bitrv2(n, a);
            cftfsub(n, a);
            rftfsub(n, a);
        } else if (n == 4) {
            cftfsub(n, a);
        }
        xr = a[0] - a[1];
        a[0] += a[1];
        for (j = 2; j < n; j += 2) {
            a[j - 1] = a[j] - a[j + 1];
            a[j] += a[j + 1];
        }
        a[n - 1] = xr;
    }
}


void ddst(int n, int isgn, double *a)
{
    void bitrv2(int n, double *a);
    void cftfsub(int n, double *a);
    void cftbsub(int n, double *a);
    void rftfsub(int n, double *a);
    void rftbsub(int n, double *a);
    void dstsub(int n, double *a);
    void dstsub4(int n, double *a);
    int j;
    double xr;
    
    if (isgn < 0) {
        xr = a[n - 1];
        for (j = n - 2; j >= 2; j -= 2) {
            a[j + 1] = -a[j] - a[j - 1];
            a[j] -= a[j - 1];
        }
        a[1] = a[0] + xr;
        a[0] -= xr;
        if (n > 4) {
            rftbsub(n, a);
            bitrv2(n, a);
            cftbsub(n, a);
        } else if (n == 4) {
            cftfsub(n, a);
        }
    }
    if (n > 4) {
        dstsub(n, a);
    } else {
        dstsub4(n, a);
    }
    if (isgn >= 0) {
        if (n > 4) {
            bitrv2(n, a);
            cftfsub(n, a);
            rftfsub(n, a);
        } else if (n == 4) {
            cftfsub(n, a);
        }
        xr = a[0] - a[1];
        a[0] += a[1];
        for (j = 2; j < n; j += 2) {
            a[j - 1] = -a[j] - a[j + 1];
            a[j] -= a[j + 1];
        }
        a[n - 1] = -xr;
    }
}


void dfct(int n, double *a)
{
    void ddct(int n, int isgn, double *a);
    void bitrv1(int n, double *a);
    int j, k, m, mh;
    double xr, xi, yr, yi, an;
    
    m = n >> 1;
    for (j = 0; j < m; j++) {
        k = n - j;
        xr = a[j] + a[k];
        a[j] -= a[k];
        a[k] = xr;
    }
    an = a[n];
    while (m >= 2) {
        ddct(m, 1, a);
        bitrv1(m, a);
        mh = m >> 1;
        xi = a[m];
        a[m] = a[0];
        a[0] = an - xi;
        an += xi;
        for (j = 1; j < mh; j++) {
            k = m - j;
            xr = a[m + k];
            xi = a[m + j];
            yr = a[j];
            yi = a[k];
            a[m + j] = yr;
            a[m + k] = yi;
            a[j] = xr - xi;
            a[k] = xr + xi;
        }
        xr = a[mh];
        a[mh] = a[m + mh];
        a[m + mh] = xr;
        m = mh;
    }
    xi = a[1];
    a[1] = a[0];
    a[0] = an + xi;
    a[n] = an - xi;
    bitrv1(n, a);
}


void dfst(int n, double *a)
{
    void ddst(int n, int isgn, double *a);
    void bitrv1(int n, double *a);
    int j, k, m, mh;
    double xr, xi, yr, yi;
    
    m = n >> 1;
    for (j = 1; j < m; j++) {
        k = n - j;
        xr = a[j] - a[k];
        a[j] += a[k];
        a[k] = xr;
    }
    a[0] = a[m];
    while (m >= 2) {
        ddst(m, 1, a);
        bitrv1(m, a);
        mh = m >> 1;
        for (j = 1; j < mh; j++) {
            k = m - j;
            xr = a[m + k];
            xi = a[m + j];
            yr = a[j];
            yi = a[k];
            a[m + j] = yr;
            a[m + k] = yi;
            a[j] = xr + xi;
            a[k] = xr - xi;
        }
        a[m] = a[0];
        a[0] = a[m + mh];
        a[m + mh] = a[mh];
        m = mh;
    }
    a[1] = a[0];
    a[0] = 0;
    bitrv1(n, a);
}


/* -------- child routines -------- */

void bitrv2(int n, double *a)
{
    int j0, k0, j1, k1, l, m, i, j, k;
    double xr, xi, yr, yi;
    
    l = n >> 2;
    m = 2;
    while (m < l) {
        l >>= 1;
        m <<= 1;
    }
    if (m == l) {
        j0 = 0;
        for (k0 = 0; k0 < m; k0 += 2) {
            k = k0;
            for (j = j0; j < j0 + k0; j += 2) {
                xr = a[j];
                xi = a[j + 1];
                yr = a[k];
                yi = a[k + 1];
                a[j] = yr;
                a[j + 1] = yi;
                a[k] = xr;
                a[k + 1] = xi;
                j1 = j + m;
                k1 = k + 2 * m;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m;
                k1 -= m;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m;
                k1 += 2 * m;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                for (i = n >> 1; i > (k ^= i); i >>= 1);
            }
            j1 = j0 + k0 + m;
            k1 = j1 + m;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            for (i = n >> 1; i > (j0 ^= i); i >>= 1);
        }
    } else {
        j0 = 0;
        for (k0 = 2; k0 < m; k0 += 2) {
            for (i = n >> 1; i > (j0 ^= i); i >>= 1);
            k = k0;
            for (j = j0; j < j0 + k0; j += 2) {
                xr = a[j];
                xi = a[j + 1];
                yr = a[k];
                yi = a[k + 1];
                a[j] = yr;
                a[j + 1] = yi;
                a[k] = xr;
                a[k + 1] = xi;
                j1 = j + m;
                k1 = k + m;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                for (i = n >> 1; i > (k ^= i); i >>= 1);
            }
        }
    }
}


void bitrv2conj(int n, double *a)
{
    int j0, k0, j1, k1, l, m, i, j, k;
    double xr, xi, yr, yi;
    
    l = n >> 2;
    m = 2;
    while (m < l) {
        l >>= 1;
        m <<= 1;
    }
    if (m == l) {
        j0 = 0;
        for (k0 = 0; k0 < m; k0 += 2) {
            k = k0;
            for (j = j0; j < j0 + k0; j += 2) {
                xr = a[j];
                xi = -a[j + 1];
                yr = a[k];
                yi = -a[k + 1];
                a[j] = yr;
                a[j + 1] = yi;
                a[k] = xr;
                a[k + 1] = xi;
                j1 = j + m;
                k1 = k + 2 * m;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m;
                k1 -= m;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m;
                k1 += 2 * m;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                for (i = n >> 1; i > (k ^= i); i >>= 1);
            }
            k1 = j0 + k0;
            a[k1 + 1] = -a[k1 + 1];
            j1 = k1 + m;
            k1 = j1 + m;
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            k1 += m;
            a[k1 + 1] = -a[k1 + 1];
            for (i = n >> 1; i > (j0 ^= i); i >>= 1);
        }
    } else {
        a[1] = -a[1];
        a[m + 1] = -a[m + 1];
        j0 = 0;
        for (k0 = 2; k0 < m; k0 += 2) {
            for (i = n >> 1; i > (j0 ^= i); i >>= 1);
            k = k0;
            for (j = j0; j < j0 + k0; j += 2) {
                xr = a[j];
                xi = -a[j + 1];
                yr = a[k];
                yi = -a[k + 1];
                a[j] = yr;
                a[j + 1] = yi;
                a[k] = xr;
                a[k + 1] = xi;
                j1 = j + m;
                k1 = k + m;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                for (i = n >> 1; i > (k ^= i); i >>= 1);
            }
            k1 = j0 + k0;
            a[k1 + 1] = -a[k1 + 1];
            a[k1 + m + 1] = -a[k1 + m + 1];
        }
    }
}


void bitrv1(int n, double *a)
{
    int j0, k0, j1, k1, l, m, i, j, k;
    double x;
    
    l = n >> 2;
    m = 1;
    while (m < l) {
        l >>= 1;
        m <<= 1;
    }
    if (m == l) {
        j0 = 0;
        for (k0 = 0; k0 < m; k0++) {
            k = k0;
            for (j = j0; j < j0 + k0; j++) {
                x = a[j];
                a[j] = a[k];
                a[k] = x;
                j1 = j + m;
                k1 = k + 2 * m;
                x = a[j1];
                a[j1] = a[k1];
                a[k1] = x;
                j1 += m;
                k1 -= m;
                x = a[j1];
                a[j1] = a[k1];
                a[k1] = x;
                j1 += m;
                k1 += 2 * m;
                x = a[j1];
                a[j1] = a[k1];
                a[k1] = x;
                for (i = n >> 1; i > (k ^= i); i >>= 1);
            }
            j1 = j0 + k0 + m;
            k1 = j1 + m;
            x = a[j1];
            a[j1] = a[k1];
            a[k1] = x;
            for (i = n >> 1; i > (j0 ^= i); i >>= 1);
        }
    } else {
        j0 = 0;
        for (k0 = 1; k0 < m; k0++) {
            for (i = n >> 1; i > (j0 ^= i); i >>= 1);
            k = k0;
            for (j = j0; j < j0 + k0; j++) {
                x = a[j];
                a[j] = a[k];
                a[k] = x;
                j1 = j + m;
                k1 = k + m;
                x = a[j1];
                a[j1] = a[k1];
                a[k1] = x;
                for (i = n >> 1; i > (k ^= i); i >>= 1);
            }
        }
    }
}


void cftfsub(int n, double *a)
{
    void cft1st(int n, double *a);
    void cftmdl(int n, int l, double *a);
    int j, j1, j2, j3, l;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    l = 2;
    if (n > 8) {
        cft1st(n, a);
        l = 8;
        while ((l << 2) < n) {
            cftmdl(n, l, a);
            l <<= 2;
        }
    }
    if ((l << 2) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            a[j2] = x0r - x2r;
            a[j2 + 1] = x0i - x2i;
            a[j1] = x1r - x3i;
            a[j1 + 1] = x1i + x3r;
            a[j3] = x1r + x3i;
            a[j3 + 1] = x1i - x3r;
        }
    } else {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = a[j] - a[j1];
            x0i = a[j + 1] - a[j1 + 1];
            a[j] += a[j1];
            a[j + 1] += a[j1 + 1];
            a[j1] = x0r;
            a[j1 + 1] = x0i;
        }
    }
}


void cftbsub(int n, double *a)
{
    void cft1st(int n, double *a);
    void cftmdl(int n, int l, double *a);
    int j, j1, j2, j3, l;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    l = 2;
    if (n > 8) {
        cft1st(n, a);
        l = 8;
        while ((l << 2) < n) {
            cftmdl(n, l, a);
            l <<= 2;
        }
    }
    if ((l << 2) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = -a[j + 1] - a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = -a[j + 1] + a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i - x2i;
            a[j2] = x0r - x2r;
            a[j2 + 1] = x0i + x2i;
            a[j1] = x1r - x3i;
            a[j1 + 1] = x1i - x3r;
            a[j3] = x1r + x3i;
            a[j3 + 1] = x1i + x3r;
        }
    } else {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = a[j] - a[j1];
            x0i = -a[j + 1] + a[j1 + 1];
            a[j] += a[j1];
            a[j + 1] = -a[j + 1] - a[j1 + 1];
            a[j1] = x0r;
            a[j1 + 1] = x0i;
        }
    }
}


void cft1st(int n, double *a)
{
    int j, kj, kr;
    double ew, wn4r, wk1r, wk1i, wk2r, wk2i, wk3r, wk3i;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    x0r = a[0] + a[2];
    x0i = a[1] + a[3];
    x1r = a[0] - a[2];
    x1i = a[1] - a[3];
    x2r = a[4] + a[6];
    x2i = a[5] + a[7];
    x3r = a[4] - a[6];
    x3i = a[5] - a[7];
    a[0] = x0r + x2r;
    a[1] = x0i + x2i;
    a[4] = x0r - x2r;
    a[5] = x0i - x2i;
    a[2] = x1r - x3i;
    a[3] = x1i + x3r;
    a[6] = x1r + x3i;
    a[7] = x1i - x3r;
    wn4r = WR5000;
    x0r = a[8] + a[10];
    x0i = a[9] + a[11];
    x1r = a[8] - a[10];
    x1i = a[9] - a[11];
    x2r = a[12] + a[14];
    x2i = a[13] + a[15];
    x3r = a[12] - a[14];
    x3i = a[13] - a[15];
    a[8] = x0r + x2r;
    a[9] = x0i + x2i;
    a[12] = x2i - x0i;
    a[13] = x0r - x2r;
    x0r = x1r - x3i;
    x0i = x1i + x3r;
    a[10] = wn4r * (x0r - x0i);
    a[11] = wn4r * (x0r + x0i);
    x0r = x3i + x1r;
    x0i = x3r - x1i;
    a[14] = wn4r * (x0i - x0r);
    a[15] = wn4r * (x0i + x0r);
    ew = M_PI_2 / n;
    kr = 0;
    for (j = 16; j < n; j += 16) {
        for (kj = n >> 2; kj > (kr ^= kj); kj >>= 1);
        wk1r = cos(ew * kr);
        wk1i = sin(ew * kr);
        wk2r = 1 - 2 * wk1i * wk1i;
        wk2i = 2 * wk1i * wk1r;
        wk3r = wk1r - 2 * wk2i * wk1i;
        wk3i = 2 * wk2i * wk1r - wk1i;
        x0r = a[j] + a[j + 2];
        x0i = a[j + 1] + a[j + 3];
        x1r = a[j] - a[j + 2];
        x1i = a[j + 1] - a[j + 3];
        x2r = a[j + 4] + a[j + 6];
        x2i = a[j + 5] + a[j + 7];
        x3r = a[j + 4] - a[j + 6];
        x3i = a[j + 5] - a[j + 7];
        a[j] = x0r + x2r;
        a[j + 1] = x0i + x2i;
        x0r -= x2r;
        x0i -= x2i;
        a[j + 4] = wk2r * x0r - wk2i * x0i;
        a[j + 5] = wk2r * x0i + wk2i * x0r;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        a[j + 2] = wk1r * x0r - wk1i * x0i;
        a[j + 3] = wk1r * x0i + wk1i * x0r;
        x0r = x1r + x3i;
        x0i = x1i - x3r;
        a[j + 6] = wk3r * x0r - wk3i * x0i;
        a[j + 7] = wk3r * x0i + wk3i * x0r;
        x0r = wn4r * (wk1r - wk1i);
        wk1i = wn4r * (wk1r + wk1i);
        wk1r = x0r;
        wk3r = wk1r - 2 * wk2r * wk1i;
        wk3i = 2 * wk2r * wk1r - wk1i;
        x0r = a[j + 8] + a[j + 10];
        x0i = a[j + 9] + a[j + 11];
        x1r = a[j + 8] - a[j + 10];
        x1i = a[j + 9] - a[j + 11];
        x2r = a[j + 12] + a[j + 14];
        x2i = a[j + 13] + a[j + 15];
        x3r = a[j + 12] - a[j + 14];
        x3i = a[j + 13] - a[j + 15];
        a[j + 8] = x0r + x2r;
        a[j + 9] = x0i + x2i;
        x0r -= x2r;
        x0i -= x2i;
        a[j + 12] = -wk2i * x0r - wk2r * x0i;
        a[j + 13] = -wk2i * x0i + wk2r * x0r;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        a[j + 10] = wk1r * x0r - wk1i * x0i;
        a[j + 11] = wk1r * x0i + wk1i * x0r;
        x0r = x1r + x3i;
        x0i = x1i - x3r;
        a[j + 14] = wk3r * x0r - wk3i * x0i;
        a[j + 15] = wk3r * x0i + wk3i * x0r;
    }
}


void cftmdl(int n, int l, double *a)
{
    int j, j1, j2, j3, k, kj, kr, m, m2;
    double ew, wn4r, wk1r, wk1i, wk2r, wk2i, wk3r, wk3i;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    m = l << 2;
    for (j = 0; j < l; j += 2) {
        j1 = j + l;
        j2 = j1 + l;
        j3 = j2 + l;
        x0r = a[j] + a[j1];
        x0i = a[j + 1] + a[j1 + 1];
        x1r = a[j] - a[j1];
        x1i = a[j + 1] - a[j1 + 1];
        x2r = a[j2] + a[j3];
        x2i = a[j2 + 1] + a[j3 + 1];
        x3r = a[j2] - a[j3];
        x3i = a[j2 + 1] - a[j3 + 1];
        a[j] = x0r + x2r;
        a[j + 1] = x0i + x2i;
        a[j2] = x0r - x2r;
        a[j2 + 1] = x0i - x2i;
        a[j1] = x1r - x3i;
        a[j1 + 1] = x1i + x3r;
        a[j3] = x1r + x3i;
        a[j3 + 1] = x1i - x3r;
    }
    wn4r = WR5000;
    for (j = m; j < l + m; j += 2) {
        j1 = j + l;
        j2 = j1 + l;
        j3 = j2 + l;
        x0r = a[j] + a[j1];
        x0i = a[j + 1] + a[j1 + 1];
        x1r = a[j] - a[j1];
        x1i = a[j + 1] - a[j1 + 1];
        x2r = a[j2] + a[j3];
        x2i = a[j2 + 1] + a[j3 + 1];
        x3r = a[j2] - a[j3];
        x3i = a[j2 + 1] - a[j3 + 1];
        a[j] = x0r + x2r;
        a[j + 1] = x0i + x2i;
        a[j2] = x2i - x0i;
        a[j2 + 1] = x0r - x2r;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        a[j1] = wn4r * (x0r - x0i);
        a[j1 + 1] = wn4r * (x0r + x0i);
        x0r = x3i + x1r;
        x0i = x3r - x1i;
        a[j3] = wn4r * (x0i - x0r);
        a[j3 + 1] = wn4r * (x0i + x0r);
    }
    ew = M_PI_2 / n;
    kr = 0;
    m2 = 2 * m;
    for (k = m2; k < n; k += m2) {
        for (kj = n >> 2; kj > (kr ^= kj); kj >>= 1);
        wk1r = cos(ew * kr);
        wk1i = sin(ew * kr);
        wk2r = 1 - 2 * wk1i * wk1i;
        wk2i = 2 * wk1i * wk1r;
        wk3r = wk1r - 2 * wk2i * wk1i;
        wk3i = 2 * wk2i * wk1r - wk1i;
        for (j = k; j < l + k; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            x0r -= x2r;
            x0i -= x2i;
            a[j2] = wk2r * x0r - wk2i * x0i;
            a[j2 + 1] = wk2r * x0i + wk2i * x0r;
            x0r = x1r - x3i;
            x0i = x1i + x3r;
            a[j1] = wk1r * x0r - wk1i * x0i;
            a[j1 + 1] = wk1r * x0i + wk1i * x0r;
            x0r = x1r + x3i;
            x0i = x1i - x3r;
            a[j3] = wk3r * x0r - wk3i * x0i;
            a[j3 + 1] = wk3r * x0i + wk3i * x0r;
        }
        x0r = wn4r * (wk1r - wk1i);
        wk1i = wn4r * (wk1r + wk1i);
        wk1r = x0r;
        wk3r = wk1r - 2 * wk2r * wk1i;
        wk3i = 2 * wk2r * wk1r - wk1i;
        for (j = k + m; j < l + (k + m); j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            x0r -= x2r;
            x0i -= x2i;
            a[j2] = -wk2i * x0r - wk2r * x0i;
            a[j2 + 1] = -wk2i * x0i + wk2r * x0r;
            x0r = x1r - x3i;
            x0i = x1i + x3r;
            a[j1] = wk1r * x0r - wk1i * x0i;
            a[j1 + 1] = wk1r * x0i + wk1i * x0r;
            x0r = x1r + x3i;
            x0i = x1i - x3r;
            a[j3] = wk3r * x0r - wk3i * x0i;
            a[j3 + 1] = wk3r * x0i + wk3i * x0r;
        }
    }
}


void rftfsub(int n, double *a)
{
    int i, i0, j, k;
    double ec, w1r, w1i, wkr, wki, wdr, wdi, ss, xr, xi, yr, yi;
    
    ec = 2 * M_PI_2 / n;
    wkr = 0;
    wki = 0;
    wdi = cos(ec);
    wdr = sin(ec);
    wdi *= wdr;
    wdr *= wdr;
    w1r = 1 - 2 * wdr;
    w1i = 2 * wdi;
    ss = 2 * w1i;
    i = n >> 1;
    for (;;) {
        i0 = i - 4 * RDFT_LOOP_DIV;
        if (i0 < 4) {
            i0 = 4;
        }
        for (j = i - 4; j >= i0; j -= 4) {
            k = n - j;
            xr = a[j + 2] - a[k - 2];
            xi = a[j + 3] + a[k - 1];
            yr = wdr * xr - wdi * xi;
            yi = wdr * xi + wdi * xr;
            a[j + 2] -= yr;
            a[j + 3] -= yi;
            a[k - 2] += yr;
            a[k - 1] -= yi;
            wkr += ss * wdi;
            wki += ss * (0.5 - wdr);
            xr = a[j] - a[k];
            xi = a[j + 1] + a[k + 1];
            yr = wkr * xr - wki * xi;
            yi = wkr * xi + wki * xr;
            a[j] -= yr;
            a[j + 1] -= yi;
            a[k] += yr;
            a[k + 1] -= yi;
            wdr += ss * wki;
            wdi += ss * (0.5 - wkr);
        }
        if (i0 == 4) {
            break;
        }
        wkr = 0.5 * sin(ec * i0);
        wki = 0.5 * cos(ec * i0);
        wdr = 0.5 - (wkr * w1r - wki * w1i);
        wdi = wkr * w1i + wki * w1r;
        wkr = 0.5 - wkr;
        i = i0;
    }
    xr = a[2] - a[n - 2];
    xi = a[3] + a[n - 1];
    yr = wdr * xr - wdi * xi;
    yi = wdr * xi + wdi * xr;
    a[2] -= yr;
    a[3] -= yi;
    a[n - 2] += yr;
    a[n - 1] -= yi;
}


void rftbsub(int n, double *a)
{
    int i, i0, j, k;
    double ec, w1r, w1i, wkr, wki, wdr, wdi, ss, xr, xi, yr, yi;
    
    ec = 2 * M_PI_2 / n;
    wkr = 0;
    wki = 0;
    wdi = cos(ec);
    wdr = sin(ec);
    wdi *= wdr;
    wdr *= wdr;
    w1r = 1 - 2 * wdr;
    w1i = 2 * wdi;
    ss = 2 * w1i;
    i = n >> 1;
    a[i + 1] = -a[i + 1];
    for (;;) {
        i0 = i - 4 * RDFT_LOOP_DIV;
        if (i0 < 4) {
            i0 = 4;
        }
        for (j = i - 4; j >= i0; j -= 4) {
            k = n - j;
            xr = a[j + 2] - a[k - 2];
            xi = a[j + 3] + a[k - 1];
            yr = wdr * xr + wdi * xi;
            yi = wdr * xi - wdi * xr;
            a[j + 2] -= yr;
            a[j + 3] = yi - a[j + 3];
            a[k - 2] += yr;
            a[k - 1] = yi - a[k - 1];
            wkr += ss * wdi;
            wki += ss * (0.5 - wdr);
            xr = a[j] - a[k];
            xi = a[j + 1] + a[k + 1];
            yr = wkr * xr + wki * xi;
            yi = wkr * xi - wki * xr;
            a[j] -= yr;
            a[j + 1] = yi - a[j + 1];
            a[k] += yr;
            a[k + 1] = yi - a[k + 1];
            wdr += ss * wki;
            wdi += ss * (0.5 - wkr);
        }
        if (i0 == 4) {
            break;
        }
        wkr = 0.5 * sin(ec * i0);
        wki = 0.5 * cos(ec * i0);
        wdr = 0.5 - (wkr * w1r - wki * w1i);
        wdi = wkr * w1i + wki * w1r;
        wkr = 0.5 - wkr;
        i = i0;
    }
    xr = a[2] - a[n - 2];
    xi = a[3] + a[n - 1];
    yr = wdr * xr + wdi * xi;
    yi = wdr * xi - wdi * xr;
    a[2] -= yr;
    a[3] = yi - a[3];
    a[n - 2] += yr;
    a[n - 1] = yi - a[n - 1];
    a[1] = -a[1];
}


void dctsub(int n, double *a)
{
    int i, i0, j, k, m;
    double ec, w1r, w1i, wkr, wki, wdr, wdi, ss, xr, xi, yr, yi;
    
    ec = M_PI_2 / n;
    wkr = 0.5;
    wki = 0.5;
    w1r = cos(ec);
    w1i = sin(ec);
    wdr = 0.5 * (w1r - w1i);
    wdi = 0.5 * (w1r + w1i);
    ss = 2 * w1i;
    m = n >> 1;
    i = 0;
    for (;;) {
        i0 = i + 2 * DCST_LOOP_DIV;
        if (i0 > m - 2) {
            i0 = m - 2;
        }
        for (j = i + 2; j <= i0; j += 2) {
            k = n - j;
            xr = wdi * a[j - 1] - wdr * a[k + 1];
            xi = wdr * a[j - 1] + wdi * a[k + 1];
            wkr -= ss * wdi;
            wki += ss * wdr;
            yr = wki * a[j] - wkr * a[k];
            yi = wkr * a[j] + wki * a[k];
            wdr -= ss * wki;
            wdi += ss * wkr;
            a[k + 1] = xr;
            a[k] = yr;
            a[j - 1] = xi;
            a[j] = yi;
        }
        if (i0 == m - 2) {
            break;
        }
        wdr = cos(ec * i0);
        wdi = sin(ec * i0);
        wkr = 0.5 * (wdr - wdi);
        wki = 0.5 * (wdr + wdi);
        wdr = wkr * w1r - wki * w1i;
        wdi = wkr * w1i + wki * w1r;
        i = i0;
    }
    xr = wdi * a[m - 1] - wdr * a[m + 1];
    a[m - 1] = wdr * a[m - 1] + wdi * a[m + 1];
    a[m + 1] = xr;
    a[m] *= wki + ss * wdr;
}


void dstsub(int n, double *a)
{
    int i, i0, j, k, m;
    double ec, w1r, w1i, wkr, wki, wdr, wdi, ss, xr, xi, yr, yi;
    
    ec = M_PI_2 / n;
    wkr = 0.5;
    wki = 0.5;
    w1r = cos(ec);
    w1i = sin(ec);
    wdr = 0.5 * (w1r - w1i);
    wdi = 0.5 * (w1r + w1i);
    ss = 2 * w1i;
    m = n >> 1;
    i = 0;
    for (;;) {
        i0 = i + 2 * DCST_LOOP_DIV;
        if (i0 > m - 2) {
            i0 = m - 2;
        }
        for (j = i + 2; j <= i0; j += 2) {
            k = n - j;
            xr = wdi * a[k + 1] - wdr * a[j - 1];
            xi = wdr * a[k + 1] + wdi * a[j - 1];
            wkr -= ss * wdi;
            wki += ss * wdr;
            yr = wki * a[k] - wkr * a[j];
            yi = wkr * a[k] + wki * a[j];
            wdr -= ss * wki;
            wdi += ss * wkr;
            a[j - 1] = xr;
            a[j] = yr;
            a[k + 1] = xi;
            a[k] = yi;
        }
        if (i0 == m - 2) {
            break;
        }
        wdr = cos(ec * i0);
        wdi = sin(ec * i0);
        wkr = 0.5 * (wdr - wdi);
        wki = 0.5 * (wdr + wdi);
        wdr = wkr * w1r - wki * w1i;
        wdi = wkr * w1i + wki * w1r;
        i = i0;
    }
    xr = wdi * a[m + 1] - wdr * a[m - 1];
    a[m + 1] = wdr * a[m + 1] + wdi * a[m - 1];
    a[m - 1] = xr;
    a[m] *= wki + ss * wdr;
}


void dctsub4(int n, double *a)
{
    int m;
    double wki, wdr, wdi, xr;
    
    wki = WR5000;
    m = n >> 1;
    if (m == 2) {
        wdr = wki * WI2500;
        wdi = wki * WR2500;
        xr = wdi * a[1] - wdr * a[3];
        a[1] = wdr * a[1] + wdi * a[3];
        a[3] = xr;
    }
    a[m] *= wki;
}


void dstsub4(int n, double *a)
{
    int m;
    double wki, wdr, wdi, xr;
    
    wki = WR5000;
    m = n >> 1;
    if (m == 2) {
        wdr = wki * WI2500;
        wdi = wki * WR2500;
        xr = wdi * a[3] - wdr * a[1];
        a[3] = wdr * a[3] + wdi * a[1];
        a[1] = xr;
    }
    a[m] *= wki;
}


/* hanning filter by T.Kondo */

void hanning(double *cdata,int nfft)
{
       int n;
       float theta, fact;
       for(n=0; n<nfft; n++)
       {
          theta=(float)(360.0*(float)n/(float)(nfft-1) - 180.0);
          theta=(float)fmod(theta,360.0) ;
          fact=(float)(0.5*(1.0+cos(theta*PIRA)));
          cdata[2*n]=fact*cdata[2*n];
          cdata[2*n+1]=fact*cdata[2*n+1];
       }
}



/* added on 2017-02-16 */


void fft_mr(double *a, double *b, int ntot, int n, int nspan, int isn)
{
/*
c      subroutine fft(a,b,ntot,n,nspan,isn)
      subroutine fftmr(a,b,ntot,n,nspan,isn)
c  Multivariate complex Fourier transform, computed in place
c    using mixed-Radix fast Fourier transform algorithm.
c  by R. C. Singleton, Stanford Research Institute, Oct. 1968
c  Arrays a and b originally hold the real and imaginary
c    components of the data, and return the real and
c    imaginary components of the resulting Fourier coefficinets.
c  Multivariate data is indexed according to the FORTRAN
c    array element successor function, without limit
c    on the number of implied multiple subscripts.
c    The subroutine is called once for each variate.
c    The calls for a multivariate transform may be in any order.
c  Ntot is the total number of complex data values.
c  N is the dimensions of the current variable.
c  Nspan/n is the spacing of consecutive data values
c    while indexing the current variable.
c  The sign of isn determines the sign of the complex
c    exponential, and the magnitude of isn is normalyy one.
c  A tri-variate transform with a(n1,n2,n3),b(n1,n2,n3)
c    is computed by
c      call fft(a,b,n1*n2*n3,n1,n1,1)
c      call fft(a,b,n1*n2*n3,n2,n1*n2,1)
c      call fft(a,b,n1*n2*n3,n3,n1*n2*n3,1)
c  For a single-variate transform,
c    ntot=n=nspan=(number of complex data values), e.g.
c      call fft(a,b,n,n,n,1)
c  The data may alternatively be stored in a single complex
c    array a, then the magnitude of isn changed to two to
c    give the correct indexing increment and a(2) used to
c    pass the initial address for the sequence of imaginary
c    values, e.g.
c      call fft(a,a(2),ntot,nspan,2)
c  Arrays at(maxf),ck(maxf),bt(maxf),sk(maxf), and np(maxp)
c    are used for temporary storage.  If the available storage
c    is insufficient, the program is terminated by a stop.
c    Maxf must be .ge. the maximum prime factor of N.
c    In addition, if the square-free portion k of n has two or
c    more prime factors, them maxp must be .ge. k-1.

      dimension a(1),b(1)

c  Array storage in nfac for a maximum of 11 factors of n.
c  If n has more than one square-free factor, the product of the
c    square-free factors must be .le. 210

      dimension nfac(11),np(209)

c  Array storage for maximum prime factor of 23

      dimension at(23),ck(23),bt(23),sk(23)
      equivalence (i,ii)
      
c  The following two constants should agree with the array dimensions.
*/
    int inc,nt,ks,kspan,kspnn,nn,jc,i,jf,m,k,j,jj,kk,kt;
    int k1,k2,k3,k4;
    double rad,s72,c72,s120,radf,sd,cd,ak,bk,c1,s1,aj,bj;
    double akp,akm,ajp,ajm,bkp,bkm,bjp,bjm,c2,s2,c3,s3,aa,bb;
    int nfac[11],np[209];
    double at[23],ck[23],bt[23],sk[23];
    double wa;
    int *ii;
    int maxf=23;
    int maxp=209;

    ii=&i;

    if(n < 2) goto exit2return;
    inc=isn;
    rad=8.0*atan(1.0);
    s72=rad/5.0;
    c72=cos(s72);
    s72=sin(s72);
    s120=sqrt(0.75);
    if(isn >= 0) goto label10;
    s72=-s72;
    s120=-s120;
    rad=-rad;
    inc=-inc;
label10: 
    nt=inc*ntot;
    ks=inc*nspan;
    kspan=ks;
    nn=nt-inc;
    jc=ks/n;
    radf=rad*(double)jc*0.5;
    i=0;
    jf=0;
/*  determine the factors of n */
    m=0;
    k=n;
    goto label20;
label15: 
    m=m+1;
    //nfac(m)=4
    nfac[m-1]=4;
    k=k/16;
label20: 
    if(k-(k/16)*16 == 0) goto label15;
    j=3;
    jj=9;
    goto label30;
label25: 
    m=m+1;
    //nfac(m)=j
    nfac[m-1]=j;
    k=k/jj;
label30: 
    //if(mod(k,jj).eq.0) go to 25
    if((k % jj) == 0) goto label25;
    j=j+2;
    //jj=j**2
    jj=j*j;
    if(jj <= k) goto label30;
    if(k > 4) goto label40;
    kt=m;
    //nfac(m+1)=k
    nfac[m]=k;
    if(k != 1) m=m+1;
    goto label80;
label40: 
    if(k-(k/4)*4 != 0) goto label50;
    m=m+1;
    //nfac(m)=2
    nfac[m-1]=2;
    k=k/4;
label50: 
    kt=m;
    j=2;
label60: 
    //if(mod(k,j).ne.0) go to 70
    if((k % j) != 0) goto label70;
    m=m+1;
    nfac[m-1]=j;
    k=k/j;
label70: 
    j=((j+1)/2)*2+1;
    if(j <= k) goto label60;
label80:
    if(kt == 0) goto label100;
    j=kt;
label90: 
    m=m+1;
    nfac[m-1]=nfac[j-1];
    j=j-1;
    if(j != 0) goto label90;
/*  compute Fourier transform */
label100: 
    sd=radf/(double)kspan;
//    //cd=2.0*sin(sd)**2
    cd=2.0*sin(sd)*sin(sd);
//    wa=sin(sd);     // 2017-02-22  time doesn't change so much
//    cd=2.0*wa*wa;   // 2017-02-22
    sd=sin(sd+sd);
    kk=1;
    i=i+1;
    if(nfac[i-1] != 2) goto label400;
/*  transform for factor of 2 (including rotation factor) */
    kspan=kspan/2;
    k1=kspan+2;
label210: 
    k2=kk+kspan;
    ak=a[k2-1];
    bk=b[k2-1];
    a[k2-1]=a[kk-1]-ak;
    b[k2-1]=b[kk-1]-bk;
    a[kk-1]=a[kk-1]+ak;
    b[kk-1]=b[kk-1]+bk;
    kk=k2+kspan;
    if(kk <= nn) goto label210;
    kk=kk-nn;
    if(kk <= jc) goto label210;
    if(kk > kspan) goto label800;
label220: 
    c1=1.0-cd;
    s1=sd;
label230: 
    k2=kk+kspan;
    ak=a[kk-1]-a[k2-1];
    bk=b[kk-1]-b[k2-1];
    a[kk-1]=a[kk-1]+a[k2-1];
    b[kk-1]=b[kk-1]+b[k2-1];
    a[k2-1]=c1*ak-s1*bk;
    b[k2-1]=s1*ak+c1*bk;
    kk=k2+kspan;
    if(kk < nt) goto label230;
    k2=kk-nt;
    c1=-c1;
    kk=k1-k2;
    if(kk > k2) goto label230;
    ak=c1-(cd*c1+sd*s1);
    s1=(sd*c1-cd*s1)+s1;

/*  The following three statements compensate for truncation */
/*    error.  If rounded arithmetic is used, substitute      */
/*     c1=ak                                                 */

    c1=0.5/(ak*ak+s1*s1)+0.5;
    s1=c1*s1;
    c1=c1*ak;
    kk=kk+jc;
    if(kk < k2) goto label230;
    k1=k1+inc+inc;
    kk=(k1-kspan)/2+jc;
    if(kk <= jc+jc) goto label220;
    goto label100;
/*  transform for factor of 3 (optional code) */
label320: 
    k1=kk+kspan;
    k2=k1+kspan;
    ak=a[kk-1];
    bk=b[kk-1];
    aj=a[k1-1]+a[k2-1];
    bj=b[k1-1]+b[k2-1];
    a[kk-1]=ak+aj;
    b[kk-1]=bk+bj;
    ak=-0.5*aj+ak;
    bk=-0.5*bj+bk;
    aj=(a[k1-1]-a[k2-1])*s120;
    bj=(b[k1-1]-b[k2-1])*s120;
    a[k1-1]=ak-bj;
    b[k1-1]=bk+aj;
    a[k2-1]=ak+bj;
    b[k2-1]=bk-aj;
    kk=k2+kspan;
    if(kk < nn) goto label320;
    kk=kk-nn;
    if(kk <= kspan) goto label320;
    goto label700;
/*  transform for factor of 4 */
label400: 
    if(nfac[i-1] != 4) goto label600;
    kspnn=kspan;
    kspan=kspan/4;
label410: 
    c1=1.0;
    s1=0.0;
label420: 
    k1=kk+kspan;
    k2=k1+kspan;
    k3=k2+kspan;
    akp=a[kk-1]+a[k2-1];
    akm=a[kk-1]-a[k2-1];
    ajp=a[k1-1]+a[k3-1];
    ajm=a[k1-1]-a[k3-1];
    a[kk-1]=akp+ajp;
    ajp=akp-ajp;
    bkp=b[kk-1]+b[k2-1];
    bkm=b[kk-1]-b[k2-1];
    bjp=b[k1-1]+b[k3-1];
    bjm=b[k1-1]-b[k3-1];
    b[kk-1]=bkp+bjp;
    bjp=bkp-bjp;
    if(isn < 0) goto label450;
    akp=akm-bjm;
    akm=akm+bjm;
    bkp=bkm+ajm;
    bkm=bkm-ajm;
    if(s1 == 0.0) goto label460;
label430: 
    a[k1-1]=akp*c1-bkp*s1;
    b[k1-1]=akp*s1+bkp*c1;
    a[k2-1]=ajp*c2-bjp*s2;
    b[k2-1]=ajp*s2+bjp*c2;
    a[k3-1]=akm*c3-bkm*s3;
    b[k3-1]=akm*s3+bkm*c3;
    kk=k3+kspan;
    if(kk <= nt) goto label420;
label440: 
    c2=c1-(cd*c1+sd*s1);
    s1=(sd*c1-cd*s1)+s1;
/*  The following three statements compensate for truncation */
/*    error.  If rounded arithmetic is used, substitute      */
/*     c1=c2                                                 */
    c1=0.5/(c2*c2+s1*s1)+0.5;
    s1=c1*s1;
    c1=c1*c2;
    //c2=c1**2-s1**2;
    c2=c1*c1-s1*s1;
    s2=2.0*c1*s1;
    c3=c2*c1-s2*s1;
    s3=c2*s1+s2*c1;
    kk=kk-nt+jc;
    if(kk <= kspan) goto label420;
    kk=kk-kspan+inc;
    if(kk <= jc) goto label410;
    if(kspan == jc) goto label800;
    goto label100;
label450: 
    akp=akm+bjm;
    akm=akm-bjm;
    bkp=bkm-ajm;
    bkm=bkm+ajm;
    if(s1 != 0.0) goto label430;
label460: 
    a[k1-1]=akp;
    b[k1-1]=bkp;
    a[k2-1]=ajp;
    b[k2-1]=bjp;
    a[k3-1]=akm;
    b[k3-1]=bkm;
    kk=k3+kspan;
    if(kk <= nt) goto label420;
    goto label440;
/*  transform for factor 5 (optional code) */
label510: 
    //c2=c72**2-s72**2
    c2=c72*c72-s72*s72;
    s2=2.0*c72*s72;
label520: 
    k1=kk+kspan;
    k2=k1+kspan;
    k3=k2+kspan;
    k4=k3+kspan;
    akp=a[k1-1]+a[k4-1];
    akm=a[k1-1]-a[k4-1];
    bkp=b[k1-1]+b[k4-1];
    bkm=b[k1-1]-b[k4-1];
    ajp=a[k2-1]+a[k3-1];
    ajm=a[k2-1]-a[k3-1];
    bjp=b[k2-1]+b[k3-1];
    bjm=b[k2-1]-b[k3-1];
    aa=a[kk-1];
    bb=b[kk-1];
    a[kk-1]=aa+akp+ajp;
    b[kk-1]=bb+bkp+bjp;
    ak=akp*c72+ajp*c2+aa;
    bk=bkp*c72+bjp*c2+bb;
    aj=akm*s72+ajm*s2;
    bj=bkm*s72+bjm*s2;
    a[k1-1]=ak-bj;
    a[k4-1]=ak+bj;
    b[k1-1]=bk+aj;
    b[k4-1]=bk-aj;
    ak=akp*c2+ajp*c72+aa;
    bk=bkp*c2+bjp*c72+bb;
    aj=akm*s2-ajm*s72;
    bj=bkm*s2-bjm*s72;
    a[k2-1]=ak-bj;
    a[k3-1]=ak+bj;
    b[k2-1]=bk+aj;
    b[k3-1]=bk-aj;
    kk=k4+kspan;
    if(kk < nn) goto label520;
    kk=kk-nn;
    if(kk <= kspan) goto label520;
    goto label700;
/*   transform for odd factors */
label600: 
    k=nfac[i-1];
    kspnn=kspan;
    kspan=kspan/k;
    if(k == 3) goto label320;
    if(k == 5) goto label510;
    if(k == jf) goto label640;
    jf=k;
    s1=rad/(double)k;
    c1=cos(s1);
    s1=sin(s1);
    if(jf > maxf) goto label998;
    ck[jf-1]=1.0;
    sk[jf-1]=0.0;
    j=1;
label630: 
    ck[j-1]=ck[k-1]*c1+sk[k-1]*s1;
    sk[j-1]=ck[k-1]*s1-sk[k-1]*c1;
    k=k-1;
    ck[k-1]=ck[j-1];
    sk[k-1]=-sk[j-1];
    j=j+1;
    if(j < k) goto label630;
label640: 
    k1=kk;
    k2=kk+kspnn;
    aa=a[kk-1];
    bb=b[kk-1];
    ak=aa;
    bk=bb;
    j=1;
    k1=k1+kspan;
label650:
    k2=k2-kspan;
    j=j+1;
    at[j-1]=a[k1-1]+a[k2-1];
    ak=at[j-1]+ak;
    bt[j-1]=b[k1-1]+b[k2-1];
    bk=bt[j-1]+bk;
    j=j+1;
    at[j-1]=a[k1-1]-a[k2-1];
    bt[j-1]=b[k1-1]-b[k2-1];
    k1=k1+kspan;
    if(k1 < k2) goto label650;
    a[kk-1]=ak;
    b[kk-1]=bk;
    k1=kk;
    k2=kk+kspnn;
    j=1;
label660:
    k1=k1+kspan;
    k2=k2-kspan;
    jj=j;
    ak=aa;
    bk=bb;
    aj=0.0;
    bj=0.0;
    k=1;
label670:
    k=k+1;
    ak=at[k-1]*ck[jj-1]+ak;
    bk=bt[k-1]*ck[jj-1]+bk;
    k=k+1;
    aj=at[k-1]*sk[jj-1]+aj;
    bj=bt[k-1]*sk[jj-1]+bj;
    jj=jj+j;
    if(jj > jf) jj=jj-jf;
    if(k < jf) goto label670;
    k=jf-j;
    a[k1-1]=ak-bj;
    b[k1-1]=bk+aj;
    a[k2-1]=ak+bj;
    b[k2-1]=bk-aj;
    j=j+1;
    if(j < k) goto label660;
    kk=kk+kspnn;
    if(kk <= nn) goto label640;
    kk=kk-nn;
    if(kk <= kspan) goto label640;
/*  multiply by rotation factor (except for factors of 2 and 4) */
label700:
    if(i == m) goto label800;
    kk=jc+1;
label710:
    c2=1.0-cd;
    s1=sd;
label720:
    c1=c2;
    s2=s1;
    kk=kk+kspan;
label730:
    ak=a[kk-1];
    a[kk-1]=c2*ak-s2*b[kk-1];
    b[kk-1]=s2*ak+c2*b[kk-1];
    kk=kk+kspnn;
    if(kk <= nt) goto label730;
    ak=s1*s2;
    s2=s1*c2+c1*s2;
    c2=c1*c2-ak;
    kk=kk-nt+kspan;
    if(kk <= kspnn) goto label730;
    c2=c1-(cd*c1+sd*s1);
    s1=s1+(sd*c1-cd*s1);
/*  The following three statements compensate for truncation */
/*    error.  If rounded arithmetic is used, they may        */
/*    be deleted                                             */
    c1=0.5/(c2*c2+s1*s1)+0.5;
    s1=c1*s1;
    c2=c1*c2;
    kk=kk-kspnn+jc;
    if(kk <= kspan) goto label720;
    kk=kk-kspan+jc+inc;
    if(kk <= jc+jc) goto label710;
    goto label100;
/*  permute the results to normal order---done in two stages */
/*  permutation for square factors of n                      */
label800:
    np[0]=ks;
    if(kt == 0) goto label890;
    k=kt+kt+1;
    if(m < k) k=k-1;
    j=1;
    np[k]=jc;
label810:
    np[j]=np[j-1]/nfac[j-1];
    np[k-1]=np[k]*nfac[j-1];
    j=j+1;
    k=k-1;
    if(j < k) goto label810;
    k3=np[k];
    kspan=np[1];
    kk=jc+1;
    k2=kspan+1;
    j=1;
    if(n != ntot) goto label850;
/*  permutation for single-variate transform (optional code) */
label820:
    ak=a[kk-1];
    a[kk-1]=a[k2-1];
    a[k2-1]=ak;
    bk=b[kk-1];
    b[kk-1]=b[k2-1];
    b[k2-1]=bk;
    kk=kk+inc;
    k2=kspan+k2;
    if(k2 < ks) goto label820;
label830: 
    k2=k2-np[j-1];
    j=j+1;
    k2=np[j]+k2;
    if(k2 > np[j-1]) goto label830;
    j=1;
label840:
    if(kk < k2) goto label820;
    kk=kk+inc;
    k2=kspan+k2;
    if(k2 < ks) goto label840;
    if(kk < ks) goto label830;
    jc=k3;
    goto label890;
/*  permutation for multivariate transfer */
label850:
    k=kk+jc;
label860:
    ak=a[kk-1];
    a[kk-1]=a[k2-1];
    a[k2-1]=ak;
    bk=b[kk-1];
    b[kk-1]=b[k2-1];
    b[k2-1]=bk;
    kk=kk+inc;
    k2=k2+inc;
    if(kk < k) goto label860;
    kk=kk+ks-jc;
    k2=k2+ks-jc;
    if(kk < nt) goto label850;
    k2=k2-nt+kspan;
    kk=kk-nt+jc;
    if(k2 < ks) goto label850;
label870:
    k2=k2-np[j-1];
    j=j+1;
    k2=np[j]+k2;
    if(k2 > np[j-1]) goto label870;
    j=1;
label880:
    if(kk < k2) goto label850;
    kk=kk+jc;
    k2=kspan+k2;
    if(k2 < ks) goto label880;
    if(kk < ks) goto label870;
    jc=k3;
label890:
    if(2*kt+1 >= m) goto exit2return;
    kspnn=np[kt];
/* permutation for square-free factors of n */
    j=m-kt;
    nfac[j]=1;
label900:
    nfac[j-1]=nfac[j-1]*nfac[j];
    j=j-1;
    if(j != kt) goto label900;
    kt=kt+1;
    nn=nfac[kt-1]-1;
    if(nn > maxp) goto label998;
    jj=0;
    j=0;
    goto label906;
label902:
    jj=jj-k2;
    k2=kk;
    k=k+1;
    kk=nfac[k-1];
label904:
    jj=kk+jj;
    if(jj >= k2) goto label902;
    np[j-1]=jj;
label906:
    k2=nfac[kt-1];
    k=kt+1;
    kk=nfac[k-1];
    j=j+1;
    if(j <= nn) goto label904;
/*  determine the permutation cycles of length greater than 1 */
    j=0;
    goto label914;
label910:
    k=kk;
    kk=np[k-1];
    np[k-1]=-kk;
    if(kk != j) goto label910;
    k3=kk;
label914:
    j=j+1;
    kk=np[j-1];
    if(kk < 0) goto label914;
    if(kk != j) goto label910;
    np[j-1]=-j;
    if(j != nn) goto label914;
    maxf=inc*maxf;
/*  reorder a and b, following the permutation cycles */
    goto label950;
label924:
    j=j-1;
    if(np[j-1] < 0) goto label924;
    jj=jc;
label926: 
    kspan=jj;
    if(jj > maxf) kspan=maxf;
    jj=jj-kspan;
    k=np[j-1];
    kk=jc*k+(*ii)+jj;
    k1=kk+kspan;
    k2=0;
label928:
    k2=k2+1;
    at[k2-1]=a[k1-1];
    bt[k2-1]=b[k1-1];
    k1=k1-inc;
    if(k1 != kk) goto label928;
label932:
    k1=kk+kspan;
    k2=k1-jc*(k+np[k-1]);
    k=-np[k-1];
label936:
    a[k1-1]=a[k2-1];
    b[k1-1]=b[k2-1];
    k1=k1-inc;
    k2=k2-inc;
    if(k1 != kk) goto label936;
    kk=k2;
    if(k != j) goto label932;
    k1=kk+kspan;
    k2=0;
label940:
    k2=k2+1;
    a[k1-1]=at[k2-1];
    b[k1-1]=bt[k2-1];
    k1=k1-inc;
    if(k1 != kk) goto label940;
    if(jj != 0) goto label926;
    if(j != 1) goto label924;
label950:
    j=k3+1;
    nt=nt-kspnn;
    *ii=nt-inc+1;
    if(nt >= 0) goto label924;
    goto exit2return;
/*  error finish, insufficient array storage */
label998:
    isn=0;
exit2return:;

}

void fft1d_mr(double *cr, double *ci, int nfft, int isn)
{
/*
   1 dimensional FFT by a Mixed-Radix method

   [in]
      cr[nfft]   --  real part of data
      ci[nfft]   --  imaginary part of data
      nfft       --  # of data (not only 2^N but also any number is OK)
      isn        --  FFT direction
                     -1: forward
                      1: inverse
   [return]
      cr[nfft]   --  real part of data after FFT
      ci[nfft]   --  imaginary part of data after FFT

*/
    int n;
    fft_mr(cr,ci,nfft,nfft,nfft,isn);
    if(isn<0){
        for(n=0; n<nfft; n++){
            cr[n]/=(double)nfft;
            ci[n]/=(double)nfft;
        }
    }

}


